#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import os
import importlib
import pkg_resources
from pprint import pformat

from collections.abc import Mapping, Sequence
from typing import Dict, Any, Union

from werkzeug.routing import Map, Rule
from werkzeug.exceptions import NotFound, MethodNotAllowed

try:
    # removed in werkzeug 2.1.0
    from werkzeug.utils import escape
except ImportError:
    from markupsafe import escape

from jinja2 import Environment, FileSystemLoader
from jinja2.sandbox import SandboxedEnvironment

from txaio import make_logger

from twisted.web import resource
from twisted.web import server
from twisted.internet.defer import succeed, inlineCallbacks

from autobahn.wamp.serializer import JsonObjectSerializer
from autobahn.wamp.types import ComponentConfig
from autobahn.wamp.exception import ApplicationError
from autobahn.twisted.wamp import ApplicationSession

from crossbar.webservice.base import RootResource, RouterWebService
from crossbar.common.checkconfig import InvalidConfigException, check_dict_args

from crossbar._util import hlid, hltype

from crossbar.worker.proxy import ProxyController
from crossbar.worker.router import RouterController

__all__ = ('RouterWebServiceWap', )


class WapResource(resource.Resource):
    """
    Twisted Web resource for WAMP Application Page web service.

    This resource uses templates loaded into a Jinja2 environment to render HTML pages with data retrieved
    from a WAMP procedure call, triggered from the original Web request.
    """

    log = make_logger()

    ser = JsonObjectSerializer()

    isLeaf = True

    def __init__(self, worker: Union[RouterController, ProxyController], config: Dict[str, Any], path: str):
        """

        :param worker: The router worker controller within this Web service is started.
        :type worker: crossbar.worker.router.RouterController

        :param config: The Web service configuration item.
        :type config: dict
        """
        resource.Resource.__init__(self)

        # remember all ctor args
        self._worker = worker
        self._config = config
        self._path = path

        # extra config
        self._server_name = 'localhost'  # FIXME
        self._realm = config.get('wamp', {}).get('realm', None)
        self._authrole = config.get('wamp', {}).get('authrole', 'anonymous')

        # the following are setup in start()
        self._jinja_env = None
        self._map_adapter = None
        self._default_session = None
        self._session_cache = None

    @inlineCallbacks
    def start(self):
        """
        Start this resource, creating a new Jinja2 rendering environment and potentially creating
        a new internal WAMP service session.
        """
        assert self._jinja_env is None
        assert self._map_adapter is None
        assert self._default_session is None
        assert self._session_cache is None

        # create a service session on the router/proxy controller: this will be used
        # to call WAMP procedures to produce data to render Jinja2 templates in HTTP requests
        assert self._worker.has_role(self._realm, self._authrole)
        if isinstance(self._worker, RouterController):
            session = ApplicationSession(ComponentConfig(realm=self._realm, extra=None))
            router = self._worker._router_session_factory._routerFactory._routers[self._realm]
            self._worker._router_session_factory.add(session, router, authrole=self._authrole)
        elif isinstance(self._worker, ProxyController):
            if not self._worker.has_realm(self._realm):
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no realm "{}" in configured routes of proxy worker'.format(self._realm))
            if not self._worker.has_role(self._realm, self._authrole):
                raise ApplicationError(
                    'crossbar.error.no_such_object',
                    'no role "{}" on realm "{}" in configured routes of proxy worker'.format(
                        self._authrole, self._realm))
            session = yield self._worker.get_service_session(self._realm, self._authrole)
            if not session or not session.is_attached():
                raise ApplicationError(
                    'crossbar.error.cannot_start',
                    'could not attach service session for HTTP bridge (role "{}" on realm "{}")'.format(
                        self._authrole, self._realm))
            assert session.realm == self._realm, 'service session: requested realm "{}", but got "{}"'.format(
                self._realm, session.realm)
            assert session.authrole == self._authrole, 'service session: requested authrole "{}", but got "{}"'.format(
                self._authrole, session.authrole)
        else:
            assert False, 'logic error: unexpected worker type {} in RouterWebServiceRestCaller.create'.format(
                type(self._worker))

        # setup Jinja2 to point to our templates folder or a package resource
        templates_config = self._config.get("templates")
        if type(templates_config) == str:
            # resolve specified template directory path relative to node directory
            templates_dir = os.path.abspath(os.path.join(self._worker.config.extra.cbdir, templates_config))
            templates_source = 'directory'
        elif type(templates_config) == dict:
            # in case we got a dict, that must contain "package" and "resource" attributes
            if 'package' not in templates_config:
                raise ApplicationError('crossbar.error.invalid_configuration',
                                       'missing attribute "resource" in WAP web service configuration')
            if 'resource' not in templates_config:
                raise ApplicationError('crossbar.error.invalid_configuration',
                                       'missing attribute "resource" in WAP web service configuration')
            try:
                importlib.import_module(templates_config['package'])
            except ImportError as e:
                emsg = 'Could not import resource {} from package {}: {}'.format(templates_config['resource'],
                                                                                 templates_config['package'], e)
                raise ApplicationError('crossbar.error.invalid_configuration', emsg)
            else:
                try:
                    # resolve template directory from package resource
                    templates_dir = os.path.abspath(
                        pkg_resources.resource_filename(templates_config['package'], templates_config['resource']))
                except Exception as e:
                    emsg = 'Could not import resource {} from package {}: {}'.format(
                        templates_config['resource'], templates_config['package'], e)
                    raise ApplicationError('crossbar.error.invalid_configuration', emsg)
            templates_source = 'package'
        else:
            raise ApplicationError(
                'crossbar.error.invalid_configuration',
                'invalid type "{}" for attribute "templates" in WAP web service configuration'.format(
                    type(templates_config)))

        # setup Jinja2 environment
        if self._config.get('sandbox', True):
            # The sandboxed environment. It works like the regular environment but tells the compiler to
            # generate sandboxed code.
            # https://jinja.palletsprojects.com/en/2.11.x/sandbox/#jinja2.sandbox.SandboxedEnvironment
            self._jinja_env = SandboxedEnvironment(loader=FileSystemLoader(templates_dir), autoescape=True)
        else:
            self._jinja_env = Environment(loader=FileSystemLoader(templates_dir), autoescape=True)

        # setup Werkzeug URL map adapter
        self._map_adapter = self._create_map_adapter(self._jinja_env, self._config, self._server_name, self._path)

        # remember default session
        self._default_session = session

        # empty session cache for cookie-based session caching
        self._session_cache = {}

        self.log.info(
            'WapResource started (realm="{realm}", authrole="{authrole}", templates_dir="{templates_dir}", templates_source="{templates_source}", jinja_env={jinja_env})',
            realm=hlid(self._realm),
            authrole=hlid(self._authrole),
            templates_dir=hlid(templates_dir),
            templates_source=hlid(templates_source),
            jinja_env=hltype(self._jinja_env.__class__))

    @staticmethod
    def _create_map_adapter(jinja_env, config, server_name, path):
        """

        :param jinja_env:
        :param config:
        :param server_name:
        :param path:
        :return:
        """
        # http://werkzeug.pocoo.org/docs/dev/routing/#werkzeug.routing.Map
        map = Map()

        # Add all our routes into 'map', note each route endpoint is a tuple of the
        # topic to call, and the template to use when rendering the results.
        for route in config.get('routes', {}):

            # compute full absolute URL of route to be added - ending in Werkzeug/Routes URL pattern
            rpath = route['path']
            _rp = []
            if path != '/':
                _rp.append(path)
            if rpath != '/':
                _rp.append(rpath)
            route_url = os.path.join('/', '/'.join(_rp))

            route_method = route.get('method', 'GET')
            assert route_method in ['GET', 'POST'], 'invalid HTTP method "{}" for route on URL "{}"'.format(
                route_method, route_url)

            route_methods = [route_method]

            # note the WAMP procedure to call and the Jinja2 template to render as HTTP response
            route_endpoint = (route['call'], jinja_env.get_template(route['render']))
            route_rule = Rule(route_url, methods=route_methods, endpoint=route_endpoint)
            map.add(route_rule)

        # https://werkzeug.palletsprojects.com/en/2.1.x/routing/#werkzeug.routing.Map.bind
        map_adapter = map.bind(server_name, '/')
        return map_adapter

    def _after_call_success(self, result, request, client_return_json):
        """
        When the WAMP call attached to the URL returns, render the WAMP result
        into a Jinja2 template and return HTML to client. Alternatively, return
        the call result as plain JSON.

        :param result: The dict returned from the WAMP procedure call.
        :param request: The HTTP request.
        :param client_return_json: Flag indicating to return plain JSON (no HTML rendering.)
        """
        try:
            if client_return_json:
                rendered = self.ser.serialize(result)
                self.log.info('WapResource successfully serialized JSON result:\n{result}', result=pformat(result))
            else:
                rendered = request.template.render(result).encode('utf8')
                self.log.info('WapResource successfully rendered HTML result: {rendered} bytes',
                              rendered=len(rendered))
        except Exception as e:
            self.log.failure()
            emsg = 'WapResource render error for WAMP result of type "{}": {}'.format(type(result), e)
            self.log.warn(emsg)
            request.setResponseCode(500)
            request.write(self._render_error(emsg, request, client_return_json))
        else:
            request.write(rendered)
        request.finish()

    def _after_call_error(self, error, request, client_return_json):
        """
        Deferred error, write out the error template and finish the request

        :param error: The current deferred error object
        :param request: The original HTTP request
        """
        self.log.error('WapResource error: {error}', error=error)
        request.setResponseCode(500)
        request.write(self._render_error(error.value.error, request, client_return_json))
        request.finish()

    def _render_error(self, message, request, client_return_json=False):
        """
        Error renderer, display a basic error message to tell the user that there
        was a problem and roughly what the problem was.

        :param message: The current error message
        :param request: The original HTTP request
        :return: HTML formatted error string
        """
        if client_return_json:
            return self.ser.serialize({'error': message})
        else:
            return """
                <html>
                    <title>API Error</title>
                    <body>
                        <h3 style="color: #f00">Crossbar WAMP Application Page Error</h3>
                        <pre>{}</pre>
                    </body>
                </html>
            """.format(escape(message)).encode('utf8')

    def render(self, request):
        """
        Initiate the rendering of a HTTP/GET request by calling a WAMP procedure, the
        resulting ``dict`` is rendered together with the specified Jinja2 template
        for this URL.

        :param request: The HTTP request.
        :returns: server.NOT_DONE_YET (special)
        """
        # https://twistedmatrix.com/documents/current/api/twisted.web.resource.Resource.html#render
        # The encoded path of the request URI (_not_ (!) including query arguments),
        full_path = request.path.decode('utf-8')

        # HTTP request method (GET, POST, ..)
        http_method = request.method.decode()
        if http_method not in ['GET', 'POST']:
            request.setResponseCode(511)
            return self._render_error(
                'Method not allowed on path "{full_path}" [werkzeug.routing.MapAdapter.match]'.format(
                    full_path=full_path), request)

        # in case of HTTP/POST, read request body as one binary string
        if http_method == 'POST' and request.content:
            content_type = request.getAllHeaders().get(b'content-type', b'application/octet-stream').decode()

            # https://stackoverflow.com/a/11549600/884770
            # http://marianoiglesias.com.ar/python/file-uploading-with-multi-part-encoding-using-twisted/
            body_data = request.content.read()
            self.log.info('POST data len = {newdata_len}', newdata_len=len(body_data))
        else:
            content_type = None
            body_data = None

        # parse and decode any query parameters
        query_args = {}
        if request.args:
            for key, values in request.args.items():
                key = key.decode()
                # we only process the first header value per key (!)
                value = values[0].decode()
                query_args[key] = value
            self.log.info('Parsed query parameters: {query_args}', query_args=query_args)

        # parse client announced accept header
        client_accept = request.getAllHeaders().get(b'accept', None)
        if client_accept:
            client_accept = client_accept.decode()

        # flag indicating the client wants to get plain JSON results (not rendered HTML)
        client_return_json = client_accept == 'application/json'

        # client cookie processing
        cookie = request.received_cookies.get(b'session_cookie')
        self.log.debug('Session Cookie is ({})'.format(cookie))
        if cookie:
            session = self._session_cache.get(cookie)
            if not session:
                # FIXME: lookup role for current session
                self.log.debug('Creating a new session for cookie ({})'.format(cookie))
                authrole = 'anonymous'
                session = ApplicationSession(ComponentConfig(realm=self._realm, extra=None))
                self._worker._router_session_factory.add(session, authrole=authrole)
                self._session_cache[cookie] = session
            else:
                self.log.debug('Using a cached session for ({})'.format(cookie))
        else:
            self.log.debug('No session cookie, falling back on default session')
            session = self._default_session

        try:
            # werkzeug.routing.MapAdapter
            # https://werkzeug.palletsprojects.com/en/2.1.x/routing/#werkzeug.routing.MapAdapter.match
            (procedure, request.template), kwargs = self._map_adapter.match(full_path,
                                                                            method=http_method,
                                                                            query_args=query_args)
            if kwargs:
                if query_args:
                    kwargs.update(query_args)
            else:
                kwargs = query_args
            self.log.info('WapResource request on path "{full_path}" mapped to call of procedure "{procedure}"',
                          full_path=full_path,
                          procedure=procedure,
                          kwargs=kwargs)

            # FIXME: how do we allow calling WAMP procedures with positional args?
            if procedure:
                self.log.info(
                    'WapResource calling procedure "{procedure}" with kwargs={kwargs} and body_data_len={body_data_len}',
                    procedure=procedure,
                    kwargs=kwargs,
                    body_data_len=len(body_data) if body_data else 0)

                # we need a session to call
                if not session:
                    self.log.error('could not call procedure - no session')
                    return self._render_error('could not call procedure - no session', request)

                if body_data:
                    if kwargs:
                        d = session.call(procedure, **kwargs, data=body_data, data_type=content_type)
                    else:
                        d = session.call(procedure, data=body_data, data_type=content_type)
                else:
                    if kwargs:
                        d = session.call(procedure, **kwargs)
                    else:
                        d = session.call(procedure)
            else:
                d = succeed({})

            d.addCallbacks(self._after_call_success,
                           self._after_call_error,
                           callbackArgs=[request, client_return_json],
                           errbackArgs=[request, client_return_json])

            return server.NOT_DONE_YET

        except NotFound:
            self.log.info('URL "{url}" not found (method={method})', url=full_path, method=http_method)
            request.setResponseCode(404)
            return self._render_error(
                'Path "{full_path}" not found [werkzeug.routing.MapAdapter.match]'.format(full_path=full_path),
                request)

        except MethodNotAllowed:
            self.log.info('method={method} not allowed on URL "{url}"', url=full_path, method=http_method)
            request.setResponseCode(511)
            return self._render_error(
                'Method not allowed on path "{full_path}" [werkzeug.routing.MapAdapter.match]'.format(
                    full_path=full_path), request)

        except Exception as e:
            self.log.info('error while processing method={method} on URL "{url}": {e}',
                          url=full_path,
                          method=http_method,
                          e=e)
            request.setResponseCode(500)
            request.write(
                self._render_error(
                    'Unknown error with path "{full_path}" [werkzeug.routing.MapAdapter.match]'.format(
                        full_path=full_path), request))
            raise


class RouterWebServiceWap(RouterWebService):
    """
    WAMP Application Page service.
    """
    @staticmethod
    def check(personality, config):
        """
        Checks the configuration item. When errors are found, an
        InvalidConfigException exception is raised.

        :param personality: The node personality class.
        :param config: The Web service configuration item.
        :raises: crossbar.common.checkconfig.InvalidConfigException
        """
        if 'type' not in config:
            raise InvalidConfigException("missing mandatory attribute 'type' in Web service configuration")

        if config['type'] != 'wap':
            raise InvalidConfigException('unexpected Web service type "{}"'.format(config['type']))

        check_dict_args(
            {
                # ID of webservice (must be unique for the web transport)
                'id': (False, [str]),

                # must be equal to "wap"
                'type': (True, [str]),

                # path to prvide to Werkzeug/Routes (eg "/test" rather than "test")
                'path': (False, [str]),

                # local directory or package+resource
                'templates': (True, [str, Mapping]),

                # create sandboxed jinja2 environment
                'sandbox': (False, [bool]),

                # Web routes
                'routes': (True, [Sequence]),

                # WAMP connection configuration
                'wamp': (True, [Mapping]),
            },
            config,
            "WAMP Application Page (WAP) service configuration: {}".format(config))

        if isinstance(config['templates'], Mapping):
            check_dict_args({
                'package': (True, [str]),
                'resource': (True, [str]),
            }, config['templates'], "templates in WAP service configuration")

        for route in config['routes']:
            check_dict_args(
                {
                    'path': (True, [str]),
                    'method': (True, [str]),
                    'call': (False, [str, type(None)]),
                    'render': (True, [str]),
                }, route, "route in WAP service configuration")

        check_dict_args({
            'realm': (True, [str]),
            'authrole': (True, [str]),
        }, config['wamp'], "wamp in WAP service configuration")

    @staticmethod
    @inlineCallbacks
    def create(transport, path: str, config: Dict[str, Any]):
        """
        Create a new WAP Web service (calling WAMP and rendering Jinja2 templates).

        :param transport: Web transport on which to add the web service.
        :param path: HTTP path on which to add the web service.
        :param config: Web service configuration.
        :return: Web service instance.
        """
        personality = transport.worker.personality
        personality.WEB_SERVICE_CHECKERS['wap'](personality, config)

        resource = WapResource(transport.worker, config, path)
        yield resource.start()
        if path == '/':
            resource = RootResource(resource, {})

        return RouterWebServiceWap(transport, path, config, resource)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from collections import deque
from typing import Optional, List, Dict, Any

from txaio import use_twisted  # noqa
from txaio import make_logger, time_ns

from autobahn.util import hltype, hlval
from autobahn.wamp.interfaces import ISession
from autobahn.wamp.types import CloseDetails, SessionDetails
from autobahn.wamp.message import Publish

from crossbar.interfaces import IRealmStore
from crossbar.router.observation import UriObservationMap

__all__ = (
    'RealmStoreMemory',
    'QueuedCall',
)


class QueuedCall(object):
    __slots__ = ('session', 'call', 'registration', 'authorization')

    def __init__(self, session, call, registration, authorization):
        self.session = session
        self.call = call
        self.registration = registration
        self.authorization = authorization


class RealmStoreMemory(object):
    """
    Memory-backed realm store.
    """

    log = make_logger()

    STORE_TYPE = 'memory'

    GLOBAL_HISTORY_LIMIT = 100
    """
    The global history limit, in case not overridden.
    """
    def __init__(self, personality, factory, config):
        """

        See the example here:

        https://github.com/crossbario/crossbar-examples/tree/master/scaling-microservices/queued

        .. code-block:: json

            "store": {
                "type": "memory",
                "limit": 1000,      // global default for limit on call queues
                "call-queue": [
                    {
                        "uri": "com.example.compute",
                        "match": "exact",
                        "limit": 10000  // procedure specific call queue limit
                    }
                ]
            }
        """
        from twisted.internet import reactor

        self._reactor = reactor
        self._personality = personality
        self._factory = factory
        self._config = config

        self._type = self._config.get('type', None)
        assert self._type == self.STORE_TYPE

        # limit to event history per subscription
        self._limit = self._config.get('limit', self.GLOBAL_HISTORY_LIMIT)

        # map of publication ID -> event dict
        self._event_store = {}

        # map of publication ID -> set of subscription IDs
        self._event_subscriptions = {}

        # map of subscription ID -> (limit, deque(of publication IDs))
        self._event_history = {}

        # map: registration.id -> deque( (session, call, registration, authorization) )
        self._queued_calls = {}

        self._running = False

        self.log.info('{func} realm store initialized (type="{stype}")',
                      stype=hlval(self._type),
                      func=hltype(self.__init__))

    def type(self) -> str:
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.type`
        """
        return self._type

    def is_running(self) -> bool:
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.is_running`
        """
        return self._running

    def start(self):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.start`
        """
        if self._running:
            raise RuntimeError('store is already running')
        else:
            self.log.info(
                '{func} starting realm store type="{stype}"',
                func=hltype(self.start),
                stype=hlval(self._type),
            )

        # currently nothing to do in stores of type "memory"
        self._running = True
        self.log.info('{func} realm store ready!', func=hltype(self.start))

    def stop(self):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.stop`
        """
        if not self._running:
            raise RuntimeError('store is not running')
        else:
            self.log.info('{func} stopping realm store', func=hltype(self.start))

        # currently nothing to do in stores of type "memory"
        self._running = False

    def store_session_joined(self, session: ISession, details: SessionDetails):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.store_session_joined`
        """
        self.log.info('{func} new session joined session={session}, details={details}',
                      func=hltype(self.store_session_joined),
                      session=session,
                      details=details)

    def store_session_left(self, session: ISession, details: CloseDetails):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.store_session_left`
        """
        self.log.info('{func} session left session={session}, details={details}',
                      func=hltype(self.store_session_left),
                      session=session,
                      details=details)

    def attach_subscription_map(self, subscription_map: UriObservationMap):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.attach_subscription_map`
        """
        for sub in self._config.get('event-history', []):
            uri = sub['uri']
            match = sub.get('match', 'exact')
            observation, was_already_observed, was_first_observer = subscription_map.add_observer(self,
                                                                                                  uri=uri,
                                                                                                  match=match)
            subscription_id = observation.id

            # for in-memory history, we just use a double-ended queue
            self._event_history[subscription_id] = (sub.get('limit', self._limit), deque())

    def store_event(self, session: ISession, publication_id: int, publish: Publish):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.store_event`
        """
        assert (publication_id not in self._event_store)
        evt = {
            'time_ns': time_ns(),
            'realm': session._realm,
            'session_id': session._session_id,
            'authid': session._authid,
            'authrole': session._authrole,
            'publication': publication_id,
            'topic': publish.topic,
            'args': publish.args,
            'kwargs': publish.kwargs
        }
        self._event_store[publication_id] = evt
        self.log.debug("Event {publication_id} stored in {store_type}-store",
                       store_type=self.STORE_TYPE,
                       publication_id=publication_id)

    def store_event_history(self, publication_id: int, subscription_id: int, receiver: ISession):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.store_event_history`
        """
        # assert(publication_id in self._event_store)
        # assert(subscription_id in self._event_history)

        if publication_id not in self._event_store:
            self.log.warn('INTERNAL WARNING: event for publication {publication_id} not in event store',
                          publication_id=publication_id)

        if subscription_id not in self._event_history:
            self.log.warn(
                'INTERNAL WARNING: subscription {subscription_id} for publication {publication_id} not in event store',
                subscription_id=subscription_id,
                publication_id=publication_id)
            return

        limit, history = self._event_history[subscription_id]

        # append event to history
        history.append(publication_id)

        if publication_id not in self._event_subscriptions:
            self._event_subscriptions[publication_id] = set()

        self._event_subscriptions[publication_id].add(subscription_id)

        self.log.debug(
            "Event {publication_id} history stored in {store_type}-store for subscription {subscription_id}",
            store_type=self.STORE_TYPE,
            publication_id=publication_id,
            subscription_id=subscription_id)

        # purge history if over limit
        if len(history) > limit:

            # remove leftmost event from history
            purged_publication_id = history.popleft()

            # remove the purged publication from event subscriptions
            self._event_subscriptions[purged_publication_id].remove(subscription_id)

            self.log.debug("Event {publication_id} purged from history for subscription {subscription_id}",
                           publication_id=purged_publication_id,
                           subscription_id=subscription_id)

            # if no more event subscriptions exist for publication, remove that too
            if not self._event_subscriptions[purged_publication_id]:
                del self._event_subscriptions[purged_publication_id]
                del self._event_store[purged_publication_id]
                self.log.debug("Event {publication_id} purged completey", publication_id=purged_publication_id)

    def get_events(self, subscription_id: int, limit: Optional[int] = None):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.get_events`
        """
        if subscription_id not in self._event_history:
            return None
        else:
            _, history = self._event_history[subscription_id]

            # at most "limit" events in reverse chronological order
            res = []
            i = -1
            if limit is None or limit > len(history):
                limit = len(history)
            for _ in range(limit):
                publication_id = history[i]
                res.append(self._event_store[publication_id])
                i -= 1
            return res

    def get_event_history(self, subscription_id: int, from_ts: int, until_ts: int) -> Optional[List[Dict[str, Any]]]:
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.get_event_history`
        """
        raise Exception("not implemented")

    def maybe_queue_call(self, session, call, registration, authorization):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.maybe_queue_call`
        """
        # FIXME: match this against the config, not just plain accept queueing!
        if registration.id not in self._queued_calls:
            self._queued_calls[registration.id] = deque()

        self._queued_calls[registration.id].append(QueuedCall(session, call, registration, authorization))

        return True

    def get_queued_call(self, registration):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.get_queued_call`
        """
        if registration.id in self._queued_calls and self._queued_calls[registration.id]:
            return self._queued_calls[registration.id][0]

    def pop_queued_call(self, registration):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.pop_queued_call`
        """
        if registration.id in self._queued_calls and self._queued_calls[registration.id]:
            return self._queued_calls[registration.id].popleft()


IRealmStore.register(RealmStoreMemory)



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import os
import uuid
from typing import Optional, List, Dict
from pprint import pformat

import numpy as np

import cfxdb.mrealm.types
from autobahn import wamp
from autobahn.wamp.exception import ApplicationError
from autobahn.wamp.types import CallDetails, PublishOptions, RegisterOptions

from crossbar._util import hl, hlid, hltype, hlval
import zlmdb
from cfxdb.mrealm import ApplicationRealm, ApplicationRealmRoleAssociation, Role, Permission, \
    WorkerGroupStatus, Principal, Credential, RouterWorkerGroupClusterPlacement, Node

import txaio

txaio.use_twisted()
from txaio import time_ns, sleep, make_logger  # noqa
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.task import LoopingCall


class ApplicationRealmMonitor(object):
    """
    Background monitor running periodically in the master node to monitor, check and apply
    necessary actions for application realms.

    The monitor is started when an application realm is started (on a router worker group already
    running on a router cluster).
    """
    log = make_logger()

    def __init__(self, manager, arealm_oid, interval=10.):
        """

        :param manager: The application realm manager that has started and is hosting this monitor.
        :type manager: :class:`crossbar.master.arealm.ApplicationRealmManager`

        :param arealm_oid: Application realm object ID (not WAMP realm name).
        :type arealm_oid: :class:`uuid.UUID`

        :param interval: Run this monitor periodically every `interval` seconds.
        :type interval: float
        """
        # the arealm manager this monitor is started from
        self._manager = manager

        # the arealm this monitor is working for
        self._arealm_oid = arealm_oid

        # the interval (in seconds) this monitor will check the arealm
        self._interval = interval

        # when this monitor is running, the looping call periodically executed
        self._loop = None

        # while self._loop is being executed (which happens every self._interval secs),
        # this flag is set
        self._check_and_apply_in_progress = False

    @property
    def is_started(self):
        """

        :return: Flag indicating whether this monitor is currently running.
        """
        return self._loop is not None and self._loop.running

    def start(self):
        """
        Start this monitor. The monitor will run in the background every `interval` seconds
        until stopped.

        .. note::
            Should the monitor iteration exceed 'interval` seconds run-time, the monitor
            will skip the next iteration(s). In other words, the monitor iteration loop
            automatically prohibits being run twice in parallel.
        :return:
        """
        assert self._loop is None

        self._loop = LoopingCall(self._check_and_apply)
        self._loop.start(self._interval)

    def stop(self):
        """
        Stop running this monitor.
        """
        assert self._loop is not None

        self._loop.stop()
        self._loop = None
        self._check_and_apply_in_progress = False

    @inlineCallbacks
    def _check_and_apply(self):
        """
        Run one iteration of the background monitor check & apply cycle.

        - _apply_routercluster_placements (L194)
        - _apply_webcluster_backends (L232)
            - _apply_webcluster_routes
        """
        if self._check_and_apply_in_progress:
            # we prohibit running the iteration multiple times concurrently. this might
            # happen when the iteration takes longer than the interval the monitor is set to
            self.log.warn(
                '{func} {action} for application realm {arealm_oid} skipped! check & apply already in progress.',
                action=hl('check & apply run skipped', color='red', bold=True),
                func=hltype(self._check_and_apply),
                arealm_oid=hlid(self._arealm_oid))
            return
        else:
            self.log.info('{func} {action} for application realm {arealm_oid} ..',
                          action=hl('check & apply run started', color='green', bold=True),
                          func=hltype(self._check_and_apply),
                          arealm_oid=hlid(self._arealm_oid))
            self._check_and_apply_in_progress = True

        # collect information on resources that ought to be running for the arealm
        with self._manager.db.begin() as txn:
            arealm = self._manager.schema.arealms[txn, self._arealm_oid]

            # the object must be in database, otherwise this is a logic error
            assert arealm

            # only process arealm when status is STATUS_STARTING or STATUS_RUNNING
            if arealm.status in [ApplicationRealm.STATUS_STARTING, ApplicationRealm.STATUS_RUNNING]:

                # collect roles associated with the arealm
                roles = []
                for _, role_oid in self._manager.schema.arealm_role_associations.select(
                        txn, from_key=(self._arealm_oid, uuid.UUID(bytes=b'\0' * 16)), return_values=False):
                    role = self._manager.schema.roles[txn, role_oid]
                    roles.append(role)

                # get the routercluster workergroup for this arealm
                if arealm.workergroup_oid:
                    workergroup = self._manager.schema.router_workergroups[txn, arealm.workergroup_oid]

                    # the object must be in database, otherwise this is a logic error
                    assert workergroup

                    # get all workergroup worker placements for this workergroup
                    workergroup_placements: List[RouterWorkerGroupClusterPlacement] = []

                    # idx_clusterplacement_by_workername: (workergroup_oid, cluster_oid, node_oid, worker_name)
                    #                                       -> placement_oid
                    for placement_oid in self._manager.schema.idx_clusterplacement_by_workername.select(
                            txn,
                            from_key=(workergroup.oid, uuid.UUID(bytes=b'\0' * 16), uuid.UUID(bytes=b'\0' * 16), ''),
                            to_key=(uuid.UUID(int=(int(workergroup.oid) + 1)), uuid.UUID(bytes=b'\0' * 16),
                                    uuid.UUID(bytes=b'\0' * 16), ''),
                            return_keys=False):
                        placement = self._manager.schema.router_workergroup_placements[txn, placement_oid]
                        workergroup_placements.append(placement)

                else:
                    self.log.warn('{func} application realm in status {status}, but no worker group associated!',
                                  func=hltype(self._check_and_apply),
                                  status=hlval(ApplicationRealm.STATUS_BY_CODE[arealm.status]))
                    self._check_and_apply_in_progress = False
                    return
            else:
                self.log.warn('{func} {action} for application realm {arealm_oid} (status is {arealm_status})',
                              action=hl('check & apply skipped', color='yellow', bold=True),
                              func=hltype(self._check_and_apply),
                              arealm_oid=hlid(self._arealm_oid),
                              arealm_status=arealm.status)
                self._check_and_apply_in_progress = False
                return

        # whenever we encounter a resource not running or in error, this flag is set False
        is_running_completely = True

        # map of nodes this router worker group runs workers on
        placement_nodes = {}

        if workergroup_placements:
            self.log.info(
                '{func} Applying {cnt_placements} worker placements for router cluster worker group {workergroup_oid}, arealm {arealm_oid}',
                func=hltype(self._check_and_apply),
                cnt_placements=hlval(len(workergroup_placements)),
                workergroup_oid=hlid(arealm.workergroup_oid),
                arealm_oid=hlid(arealm.oid))

            with self._manager.gdb.begin() as txn:
                for node in self._manager.gschema.nodes.select(txn, return_keys=False):
                    placement_nodes[node.oid] = node

            # apply routercluster worker placements for nodes/workers involved
            success = yield self._apply_routercluster_placements(arealm, workergroup, workergroup_placements,
                                                                 placement_nodes)
            if not success:
                is_running_completely = False
        else:
            self.log.warn(
                '{func} no placements configured for router cluster worker group {workergroup_oid} and arealm {arealm_oid}!',
                func=hltype(self._check_and_apply),
                workergroup_oid=hlid(arealm.workergroup_oid),
                arealm_oid=hlid(arealm.oid))
            is_running_completely = False

        # if the application realm has a (frontend) webcluster assigned (to accept incoming client
        # connections), setup proxy backend connections and routes to the router workers in the
        # router worker group of the application realm
        if arealm and arealm.webcluster_oid and workergroup_placements:
            # get list of nodes/workers for the webcluster, eg:
            #
            # [('309fa904-9f32-47b9-901c-33ec78b0c582', 'cpw-edc6ba18-0'),
            #  ('309fa904-9f32-47b9-901c-33ec78b0c582', 'cpw-edc6ba18-1'),
            #  ('4f19d9e0-4b28-45fc-bbb0-8e46e6d79376', 'cpw-edc6ba18-0'),
            #  ('4f19d9e0-4b28-45fc-bbb0-8e46e6d79376', 'cpw-edc6ba18-1')]
            #
            wc_workers = self._manager._session._webcluster_manager.get_webcluster_workers(arealm.webcluster_oid,
                                                                                           filter_online=True)
            self.log.debug('{func} Ok, found {cnt_workers} running web cluster workers ..',
                           func=hltype(self._check_and_apply),
                           cnt_workers=len(wc_workers) if wc_workers else 0)

            # on each webcluster worker, setup backend connections and routes to all router workers of
            # the router worker group of this application realm
            for wc_node_oid, wc_worker_id in wc_workers:
                success = yield self._apply_webcluster_connections(wc_node_oid, wc_worker_id, workergroup_placements,
                                                                   placement_nodes, arealm)
                if not success:
                    is_running_completely = False
        else:
            self.log.warn('{func} application realm in status {status}, but no web cluster associated!',
                          func=hltype(self._check_and_apply),
                          status=hlval(ApplicationRealm.STATUS_BY_CODE[arealm.status]))
            is_running_completely = False

        # if the status is STATUS_STARTING and we have indeed completed startup in this iteration,
        # update the status in the database and publish an event
        if arealm.status in [ApplicationRealm.STATUS_STARTING] and is_running_completely:
            with self._manager.db.begin(write=True) as txn:
                arealm = self._manager.schema.arealms[txn, self._arealm_oid]
                arealm.status = ApplicationRealm.STATUS_RUNNING
                arealm.changed = time_ns()
                self._manager.schema.arealms[txn, arealm.oid] = arealm

            arealm_started = {
                'oid': str(arealm.oid),
                'status': ApplicationRealm.STATUS_BY_CODE[arealm.status],
                'changed': arealm.changed,
            }
            yield self._manager._session.publish('{}.on_arealm_started'.format(self._manager._prefix),
                                                 arealm_started,
                                                 options=self._manager._PUBOPTS)

        if is_running_completely:
            color = 'green'
            action = 'check & apply run completed successfully'
        else:
            color = 'red'
            action = 'check & apply run finished with problems left'

        self._check_and_apply_in_progress = False
        self.log.info('{func} {action} for application realm {arealm_oid}!',
                      action=hl(action, color=color, bold=True),
                      func=hltype(self._check_and_apply),
                      arealm_oid=hlid(self._arealm_oid))

    @inlineCallbacks
    def _apply_webcluster_connections(self, wc_node_oid, wc_worker_id, workergroup_placements, placement_nodes,
                                      arealm):
        """
        For given webcluster worker ``(wc_node_oid, wc_worker_id)``, setup backend connections and routes
        to all router workers ``workergroup_placements`` of the router worker group of the given
        application realm ``arealm``.
        """
        is_running_completely = True

        # iterate over all router worker placements in the router cluster worker group.
        # for each placement on the given (wc_node_oid, wc_worker_id) webcluster worker, we need to setup:
        #   - a connection and
        #   - routes
        for placement in workergroup_placements:
            self.log.debug('{func} applying router cluster worker group placement:\n{placement}',
                           func=hltype(self._apply_webcluster_connections),
                           placement=pformat(placement.marshal()))

            # the router worker this placement targets
            node_oid = placement.node_oid
            worker_name = placement.worker_name
            tcp_listening_port = placement.tcp_listening_port

            node_authid = placement_nodes[node_oid].authid
            node_cluster_ip = placement_nodes[node_oid].cluster_ip

            # we will name the connection on our proxy worker along the target node/worker
            connection_id = 'cnc_{}_{}'.format(node_authid, worker_name)

            # check if a connection with the respective name is already running
            try:
                connection = yield self._manager._session.call(
                    'crossbarfabriccenter.remote.proxy.get_proxy_connection', str(wc_node_oid), wc_worker_id,
                    connection_id)
            except ApplicationError as e:
                if e.error != 'crossbar.error.no_such_object':
                    # anything but "no_such_object" is unexpected (and fatal)
                    raise
                is_running_completely = False
                connection = None
                self.log.warn(
                    '{func} No connection {connection_id} currently running for web cluster (proxy) worker {wc_worker_id} on node {wc_node_oid}: starting backend connection ..',
                    func=hltype(self._apply_webcluster_connections),
                    wc_node_oid=hlid(wc_node_oid),
                    wc_worker_id=hlid(wc_worker_id),
                    connection_id=hlid(connection_id))
            else:
                self.log.debug(
                    '{func} Ok, connection {connection_id} already running on web cluster (proxy) worker {wc_worker_id} on node {wc_node_oid}',
                    func=hltype(self._apply_webcluster_connections),
                    wc_node_oid=hlid(wc_node_oid),
                    wc_worker_id=hlid(wc_worker_id),
                    connection_id=hlid(connection_id))

            # if no connection is running, and if the target placement has a TCP port configured,
            # start a new connection on the proxy worker
            #
            # see: :class:`crossbar.worker.proxy.ProxyConnection`
            #
            if not connection and tcp_listening_port:
                # transport configuration for proxy worker -> router worker connections
                config = {
                    "transport": {
                        "type": "rawsocket",
                        "endpoint": {
                            "type": "tcp",
                            # we assume the authid for the node was set to the hostname of that node, and
                            # that the node is reachable using this hostname from other nodes (eg inside a
                            # private network)
                            "host": node_cluster_ip,
                            "port": tcp_listening_port
                        },
                        "url": "ws://{}".format(node_authid),
                        "serializer": "cbor"
                    },
                    "auth": {
                        # must use cryptosign-proxy, NOT anonymous-proxy, since we run
                        # over TCP, not UDS, and to IP addresses on different hosts
                        "cryptosign-proxy": {
                            "type": "static"
                        }
                    }
                }
                connection = yield self._manager._session.call(
                    'crossbarfabriccenter.remote.proxy.start_proxy_connection', str(wc_node_oid), wc_worker_id,
                    connection_id, config)

                self.log.info('{func} Proxy backend connection started:\n{connection}',
                              func=hltype(self._apply_webcluster_connections),
                              connection=connection)

            # if by now we do have a connection on the proxy worker, setup all routes (for the arealm)
            # using the connection
            #
            # see: :class:`crossbar.worker.proxy.ProxyRoute`
            #
            if connection:
                # proxy routes use "realm name" as key (not a synthetic ID)
                realm_name = arealm.name
                routes = []

                try:
                    routes = yield self._manager._session.call(
                        'crossbarfabriccenter.remote.proxy.list_proxy_realm_routes', str(wc_node_oid), wc_worker_id,
                        realm_name)
                except ApplicationError as e:
                    if e.error != 'crossbar.error.no_such_object':
                        # anything but "no_such_object" is unexpected (and fatal)
                        raise
                    is_running_completely = False
                    self.log.warn(
                        '{func} No route for realm "{realm_name}" currently running for web cluster (proxy) worker {wc_worker_id} on node {wc_node_oid}: starting backend route ..',
                        func=hltype(self._apply_webcluster_connections),
                        wc_node_oid=hlid(wc_node_oid),
                        wc_worker_id=hlid(wc_worker_id),
                        realm_name=hlval(realm_name))
                else:
                    self.log.debug(
                        '{func} Ok, route for realm "{realm_name}" already running on web cluster (proxy) worker {wc_worker_id} on node {wc_node_oid}',
                        func=hltype(self._apply_webcluster_connections),
                        wc_node_oid=hlid(wc_node_oid),
                        wc_worker_id=hlid(wc_worker_id),
                        realm_name=hlval(realm_name))

                if (not routes or len(routes) != len(workergroup_placements)):
                    new_routes, _is_running_completely = yield self._apply_webcluster_routes(
                        workergroup_placements,
                        wc_node_oid,
                        wc_worker_id,
                        arealm,
                        placement_nodes,
                    )
                    routes.extend(new_routes)
                    if is_running_completely:
                        is_running_completely = _is_running_completely

        return is_running_completely

    @inlineCallbacks
    def _apply_webcluster_routes(self, workergroup_placements, wc_node_oid, wc_worker_id, arealm,
                                 placement_nodes_keys):
        is_running_completely = True
        realm_name = arealm.name
        try:
            routes = yield self._manager._session.call(
                'crossbarfabriccenter.remote.proxy.list_proxy_realm_routes',
                str(wc_node_oid),
                wc_worker_id,
                realm_name,
            )

        except ApplicationError as e:
            if e.error != 'crossbar.error.no_such_object':
                # anything but "no_such_object" is unexpected (and fatal)
                raise
            is_running_completely = False
            self.log.info(
                '{func} No routes for realm "{realm_name}" currently running for web cluster (proxy) worker {wc_worker_id} on node {wc_node_oid}: starting backend routes ..',
                func=hltype(self._apply_webcluster_routes),
                wc_node_oid=hlid(wc_node_oid),
                wc_worker_id=hlid(wc_worker_id),
                realm_name=hlval(realm_name))
            routes = []

        # FIXME: proxy routes for all realms X roles
        for placement in workergroup_placements:

            # can placement.node_oid *not* be in this map?
            node_authid = placement_nodes_keys[placement.node_oid].authid
            connection_id = 'cnc_{}_{}'.format(node_authid, placement.worker_name)

            # config = {'anonymous': connection_id}
            config = {}
            from_key = (arealm.oid, uuid.UUID(bytes=b'\x00' * 16))
            to_key = (uuid.UUID(int=(int(arealm.oid) + 1)), uuid.UUID(bytes=b'\x00' * 16))
            with self._manager.db.begin() as txn:
                for _, role_oid in self._manager.schema.arealm_role_associations.select(txn,
                                                                                        from_key=from_key,
                                                                                        to_key=to_key,
                                                                                        return_values=False):
                    role = self._manager.schema.roles[txn, role_oid]
                    config[role.name] = connection_id

            # existing = any(route['config'].get(role.name, None) == connection_id for route in routes)

            try:
                route = yield self._manager._session.call('crossbarfabriccenter.remote.proxy.start_proxy_realm_route',
                                                          str(wc_node_oid), wc_worker_id, realm_name, config)
            except Exception as e:
                self.log.error('Proxy route failed: {e}', e=e)
            else:
                self.log.info(
                    '{func} Proxy backend route started:\n{route}',
                    func=hltype(self._apply_webcluster_routes),
                    route=route,
                )
                routes.append(route)
        returnValue((routes, is_running_completely))

    @inlineCallbacks
    def _apply_routercluster_placements(self, arealm: cfxdb.mrealm.ApplicationRealm,
                                        workergroup: cfxdb.mrealm.RouterWorkerGroup,
                                        workergroup_placements: List[RouterWorkerGroupClusterPlacement],
                                        placement_nodes: Dict[uuid.UUID, Node]):
        """
        Apply worker placements for workergroup of routercluster.

        :param arealm: Application realm to process placements for.
        :param workergroup: Router worker group to process placements for.
        :param workergroup_placements: List of placements.
        :param placement_nodes: Map of node object IDs to pair of node public key and authid.
        """
        # this flag will remain true as long as we could process all placements successfully
        is_running_completely = True

        # I) iterate over all placements sequentially
        for placement in workergroup_placements:
            self.log.info('{func} Applying router cluster worker group placement:\n{placement}',
                          func=hltype(self._apply_routercluster_placements),
                          placement=pformat(placement.marshal()))

            # place the worker on this node and (router) worker
            node_oid = placement.node_oid
            worker_name = placement.worker_name

            # get run-time information for the node (as maintained here in our master view of the external world)
            # instance of crossbar.master.mrealm.controller.Node
            node = self._manager._session.nodes.get(str(node_oid), None)

            # the node must be found and must be currently online for us to manage it
            if node and node.status == 'online':
                node_authid = placement_nodes[node_oid].authid

                self.log.debug('{func} Ok, router cluster node "{node_authid}" ({node_oid}) is running!',
                               func=hltype(self._apply_routercluster_placements),
                               node_authid=hlid(node_authid),
                               node_oid=hlid(node_oid))

                # II.1) get worker run-time information (obtained by calling into the live node)
                worker = None
                try:
                    worker = yield self._manager._session.call('crossbarfabriccenter.remote.node.get_worker',
                                                               str(node_oid), worker_name)
                except ApplicationError as e:
                    if e.error != 'crossbar.error.no_such_worker':
                        # anything but "no_such_worker" is unexpected (and fatal)
                        raise
                    self.log.warn(
                        '{func} No router cluster worker {worker_name} currently running on node {node_oid}: starting worker ..',
                        func=hltype(self._apply_routercluster_placements),
                        node_oid=hlid(node_oid),
                        worker_name=hlid(worker_name))
                except:
                    self.log.failure()
                    raise
                else:
                    self.log.debug(
                        '{func} Ok, router cluster worker {worker_name} already running on node {node_oid}!',
                        func=hltype(self._apply_routercluster_placements),
                        node_oid=hlid(node_oid),
                        worker_name=hlid(worker_name))

                # II.2) if there isn't a worker running (with worker ID as we expect) already, start a new router worker
                if not worker:
                    worker_options = {
                        'env': {
                            'inherit': ['PYTHONPATH']
                        },
                        'title': 'Managed router worker {}'.format(worker_name),
                        'extra': {}
                    }
                    try:
                        worker_started = yield self._manager._session.call(
                            'crossbarfabriccenter.remote.node.start_worker', str(node_oid), worker_name, 'router',
                            worker_options)
                        worker = yield self._manager._session.call('crossbarfabriccenter.remote.node.get_worker',
                                                                   str(node_oid), worker_name)
                        self.log.info(
                            '{func} Router cluster worker {worker_name} started on node {node_oid} [{worker_started}]',
                            func=hltype(self._apply_routercluster_placements),
                            node_oid=hlid(node_oid),
                            worker_name=hlid(worker['id']),
                            worker_started=worker_started)
                    except:
                        self.log.failure()
                        is_running_completely = False

                # we can only continue with transport(s) when we now have a worker started already
                if worker:

                    transport_id = 'tnp_{}'.format(worker_name)
                    transport = None

                    # III.1) get transport run-time information (obtained by calling into the live node)
                    try:
                        transport = yield self._manager._session.call(
                            'crossbarfabriccenter.remote.router.get_router_transport', str(node_oid), worker_name,
                            transport_id)
                    except ApplicationError as e:
                        if e.error != 'crossbar.error.no_such_object':
                            # anything but "no_such_object" is unexpected (and fatal)
                            raise
                        self.log.info(
                            '{func} No Transport {transport_id} currently running for Web cluster worker {worker_name}: starting transport ..',
                            func=hltype(self._apply_routercluster_placements),
                            worker_name=hlid(worker_name),
                            transport_id=hlid(transport_id))
                    else:
                        self.log.debug(
                            '{func} Ok, transport {transport_id} already running on Web cluster worker {worker_name}',
                            func=hltype(self._apply_routercluster_placements),
                            worker_name=hlid(worker_name),
                            transport_id=hlid(transport_id))

                    # III.2) if there isn't a transport started (with transport ID as we expect) already,
                    # start a new transport
                    if not transport:

                        # FIXME: allow to configure transport type TCP vs UDS
                        USE_UDS = False

                        if USE_UDS:
                            # https://serverfault.com/a/641387/117074
                            UNIX_PATH_MAX = 108
                            transport_path = os.path.abspath('{}.sock'.format(transport_id))
                            if len(transport_path) > UNIX_PATH_MAX:
                                raise RuntimeError(
                                    'unix domain socket path too long! was {}, but maximum is {}'.format(
                                        len(transport_path), UNIX_PATH_MAX))
                            transport_config = {
                                'id': transport_id,
                                'type': 'rawsocket',
                                'endpoint': {
                                    'type': 'unix',
                                    'path': transport_path
                                },
                                'options': {
                                    "max_message_size": 1048576
                                },
                                'serializers': ['cbor'],
                                'auth': {
                                    # use anonymous-proxy authentication for UDS based connections (on localhost only)
                                    'anonymous-proxy': {
                                        'type': 'static'
                                    }
                                }
                            }
                        else:
                            principals = {}
                            all_pubkeys = [node.pubkey for node in placement_nodes.values()]
                            for node in placement_nodes.values():
                                principal = {
                                    'realm': arealm.name,
                                    'role': 'rlink',
                                    'authorized_keys': all_pubkeys,
                                }
                                principals[node.authid] = principal

                            transport_config = {
                                'id': transport_id,
                                'type': 'rawsocket',
                                'endpoint': {
                                    'type': 'tcp',
                                    # let the router worker auto-assign a listening port from this range
                                    'portrange': [10000, 10100]
                                },
                                'options': {
                                    "max_message_size": 1048576
                                },
                                'serializers': ['cbor'],
                                'auth': {
                                    # use cryptosign-proxy authentication for TCP based connections
                                    'cryptosign-proxy': {
                                        'type': 'static',
                                        'principals': principals
                                    }
                                }
                            }

                        try:
                            transport_started = yield self._manager._session.call(
                                'crossbarfabriccenter.remote.router.start_router_transport', str(node_oid),
                                worker_name, transport_id, transport_config)
                            transport = yield self._manager._session.call(
                                'crossbarfabriccenter.remote.router.get_router_transport', str(node_oid), worker_name,
                                transport_id)
                            self.log.info(
                                '{func} Transport {transport_id} started on router cluster worker {worker_name} [{transport_started}]',
                                func=hltype(self._apply_routercluster_placements),
                                worker_name=hlid(worker_name),
                                transport_id=hlid(transport_id),
                                transport_started=transport_started)
                        except:
                            self.log.failure()
                            is_running_completely = False
                        else:
                            # when a new transport was started with an auto-assigned portrange, grab the
                            # actual TCP listening port that was selected on the target node
                            tcp_listening_port = transport_started['config']['endpoint']['port']

                            with self._manager.db.begin(write=True) as txn:
                                placement = self._manager.schema.router_workergroup_placements[txn, placement.oid]
                                placement.changed = time_ns()
                                placement.status = WorkerGroupStatus.RUNNING
                                placement.tcp_listening_port = tcp_listening_port
                                self._manager.schema.router_workergroup_placements[txn, placement.oid] = placement

                            self.log.info('{func} Ok, placement {placement_oid} updated:\n{placement}',
                                          func=hltype(self._apply_routercluster_placements),
                                          placement_oid=hlid(placement.oid),
                                          placement=placement)

                    # IV.1) get arealm run-time information (obtained by calling into the live node)

                    runtime_realm_id = 'rlm_{}'.format(str(arealm.oid)[:8])
                    running_arealm = None

                    try:
                        running_arealm = yield self._manager._session.call(
                            'crossbarfabriccenter.remote.router.get_router_realm', str(node_oid), worker_name,
                            runtime_realm_id)
                    except ApplicationError as e:
                        if e.error != 'crossbar.error.no_such_object':
                            # anything but "no_such_object" is unexpected (and fatal)
                            raise
                        self.log.info(
                            '{func} No application realm {runtime_realm_id} currently running for router cluster worker {worker_name}: starting application realm ..',
                            func=hltype(self._apply_routercluster_placements),
                            worker_name=hlid(worker_name),
                            runtime_realm_id=hlid(runtime_realm_id))
                    else:
                        self.log.debug(
                            '{func} Ok, application realm {runtime_realm_id} already running on router cluster worker {worker_name}',
                            func=hltype(self._apply_routercluster_placements),
                            worker_name=hlid(worker_name),
                            runtime_realm_id=hlid(runtime_realm_id))

                    # IV.2) if there isn't an arealm started (with realm ID as we expect) already,
                    # start a new arealm
                    if not running_arealm:
                        realm_config = {
                            "name":
                            arealm.name,

                            # built-in (reserved) roles
                            "roles": [{
                                "name":
                                "rlink",
                                "permissions": [{
                                    "uri": "",
                                    "match": "prefix",
                                    "allow": {
                                        "call": True,
                                        "register": True,
                                        "publish": True,
                                        "subscribe": True
                                    },
                                    "disclose": {
                                        "caller": True,
                                        "publisher": True
                                    },
                                    "cache": True
                                }]
                            }]
                        }
                        try:
                            # start the application realm on the remote node worker
                            realm_started = yield self._manager._session.call(
                                'crossbarfabriccenter.remote.router.start_router_realm', str(node_oid), worker_name,
                                runtime_realm_id, realm_config)
                            running_arealm = yield self._manager._session.call(
                                'crossbarfabriccenter.remote.router.get_router_realm', str(node_oid), worker_name,
                                runtime_realm_id)
                            self.log.info(
                                '{func} Application realm {runtime_realm_id} started on router cluster worker {worker_name} [{realm_started}]',
                                func=hltype(self._apply_routercluster_placements),
                                worker_name=hlid(worker_name),
                                runtime_realm_id=hlid(running_arealm['id']),
                                realm_started=realm_started)
                        except:
                            self.log.failure()
                            is_running_completely = False
                        else:
                            # start all built-in (reserved) roles on the remote node worker
                            i = 1
                            for role in realm_config['roles']:
                                runtime_role_id = 'rle_{}_builtin_{}'.format(str(arealm.oid)[:8], i)
                                role_started = yield self._manager._session.call(
                                    'crossbarfabriccenter.remote.router.start_router_realm_role', str(node_oid),
                                    worker_name, runtime_realm_id, runtime_role_id, role)
                                running_role = yield self._manager._session.call(
                                    'crossbarfabriccenter.remote.router.get_router_realm_role', str(node_oid),
                                    worker_name, runtime_realm_id, runtime_role_id)
                                self.log.info(
                                    '{func} Application realm role {runtime_role_id} started on router cluster worker {worker_name} [{role_started}]',
                                    func=hltype(self._apply_routercluster_placements),
                                    worker_name=hlid(worker_name),
                                    runtime_role_id=hlid(running_role['id']),
                                    role_started=role_started)
                                i += 1

                    if running_arealm:
                        # start all roles defined in the realm configuration on the remote node worker
                        from_key = (arealm.oid, uuid.UUID(bytes=b'\x00' * 16))
                        to_key = (uuid.UUID(int=(int(arealm.oid) + 1)), uuid.UUID(bytes=b'\x00' * 16))

                        with self._manager.db.begin() as txn:
                            for _, role_oid in self._manager.schema.arealm_role_associations.select(
                                    txn, from_key=from_key, to_key=to_key, return_values=False):
                                role = self._manager.schema.roles[txn, role_oid]

                                # make sure role name is not reserved
                                assert role.name not in [
                                    'rlink'
                                ], 'use of reserved role name "rlink" in role {}'.format(role_oid)

                                runtime_role_id = 'rle_{}'.format(str(role.oid)[:8])
                                try:
                                    running_role = yield self._manager._session.call(
                                        'crossbarfabriccenter.remote.router.get_router_realm_role', str(node_oid),
                                        worker_name, runtime_realm_id, runtime_role_id)
                                except ApplicationError as e:
                                    if e.error != 'crossbar.error.no_such_object':
                                        # anything but "no_such_object" is unexpected (and fatal)
                                        raise
                                    self.log.info(
                                        '{func} No role {runtime_role_id} currently running for router cluster worker {worker_name}: starting role ..',
                                        func=hltype(self._apply_routercluster_placements),
                                        worker_name=hlid(worker_name),
                                        runtime_role_id=hlid(runtime_role_id))

                                    permissions = []
                                    from_key2 = (role.oid, '')
                                    to_key2 = (uuid.UUID(int=(int(role.oid) + 1)), '')
                                    for permission_oid in self._manager.schema.idx_permissions_by_uri.select(
                                            txn, from_key=from_key2, to_key=to_key2, return_keys=False):
                                        permission = self._manager.schema.permissions[txn, permission_oid]
                                        permissions.append({
                                            'uri':
                                            permission.uri,
                                            'match':
                                            Permission.MATCH_TYPES_TOSTR[permission.match]
                                            if permission.match else None,
                                            'allow': {
                                                'call': permission.allow_call or False,
                                                'register': permission.allow_register or False,
                                                'publish': permission.allow_publish or False,
                                                'subscribe': permission.allow_subscribe or False
                                            },
                                            'disclose': {
                                                'caller': permission.disclose_caller or False,
                                                'publisher': permission.disclose_publisher or False
                                            },
                                            'cache':
                                            permission.cache or False
                                        })

                                    runtime_role_config = {'name': role.name, 'permissions': permissions}

                                    role_started = yield self._manager._session.call(
                                        'crossbarfabriccenter.remote.router.start_router_realm_role', str(node_oid),
                                        worker_name, runtime_realm_id, runtime_role_id, runtime_role_config)

                                    self.log.info(
                                        '{func} Application realm role {runtime_role_id} started on router cluster worker {worker_name} [{role_started}]',
                                        func=hltype(self._apply_routercluster_placements),
                                        worker_name=hlid(worker_name),
                                        runtime_role_id=hlid(runtime_role_id),
                                        role_started=role_started)
                                else:
                                    self.log.debug(
                                        '{func} Ok, role {runtime_role_id} already running for router cluster worker {worker_name} [{running_role}].',
                                        func=hltype(self._apply_routercluster_placements),
                                        worker_name=hlid(worker_name),
                                        runtime_role_id=hlid(runtime_role_id),
                                        running_role=running_role)

                    # IV.3) if we have a running application realm by now, start router-to-router links
                    # between this worker, and every other worker in this router worker group
                    if running_arealm:
                        for other_placement in workergroup_placements:
                            other_node_oid = placement.node_oid
                            other_worker_name = other_placement.worker_name
                            assert other_node_oid
                            assert other_worker_name

                            with self._manager.gdb.begin() as txn2:
                                other_node = self._manager.gschema.nodes[txn2, other_node_oid]

                            self.log.info(
                                '{func} Rlink other node worker is on node {other_node_oid}, worker {other_worker_name}, cluster_ip {cluster_ip}:\n{other_node}',
                                other_node_oid=hlid(other_node_oid),
                                other_worker_name=hlid(other_worker_name),
                                cluster_ip=hlval(other_node.cluster_ip) if other_node else None,
                                other_node=pformat(other_node.marshal()) if other_node else None,
                                func=hltype(self._apply_routercluster_placements))
                            assert other_node

                            # don't create rlinks back to a router worker itself (only all _other_
                            # router workers in the same router worker group)
                            if other_node_oid != node_oid or other_worker_name != worker_name:
                                self.log.debug(
                                    '{func} Verifying rlink from {node_oid} / {worker_name} to {other_node_oid} / {other_worker_name} ..',
                                    func=hltype(self._apply_routercluster_placements),
                                    node_oid=hlid(node_oid),
                                    worker_name=hlid(worker_name),
                                    other_node_oid=hlid(other_node_oid),
                                    other_worker_name=hlid(other_worker_name))

                                # get run-time information for the node (as maintained here in our master view of the external world)
                                # instance of crossbar.master.mrealm.controller.Node
                                other_node_status = self._manager._session.nodes.get(str(other_node_oid), None)

                                if other_node_status and other_node_status.status == 'online':

                                    # get worker run-time information (obtained by calling into the live node)
                                    worker = None
                                    try:
                                        worker = yield self._manager._session.call(
                                            'crossbarfabriccenter.remote.node.get_worker', str(other_node_oid),
                                            other_worker_name)
                                    except ApplicationError as e:
                                        if e.error != 'crossbar.error.no_such_worker':
                                            # anything but "no_such_worker" is unexpected (and fatal)
                                            raise
                                    except:
                                        self.log.failure()
                                        raise
                                    else:
                                        self.log.debug(
                                            '{func} Ok, rlink target router worker {worker_name} is running on node {node_oid}!',
                                            func=hltype(self._check_and_apply),
                                            node_oid=hlid(other_node_oid),
                                            worker_name=hlid(other_worker_name))

                                    runtime_rlink_id = 'rlk_{}_{}_{}_{}'.format(
                                        str(arealm.oid)[:8], worker_name,
                                        str(other_node_oid)[:8], other_worker_name)
                                    running_rlink = None
                                    realm_name = arealm.name

                                    try:
                                        running_rlink = yield self._manager._session.call(
                                            'crossbarfabriccenter.remote.router.get_router_realm_link',
                                            str(other_node_oid), other_worker_name, runtime_realm_id, runtime_rlink_id)
                                    except ApplicationError as e:
                                        if e.error not in [
                                                'crossbar.error.no_such_object', 'wamp.error.no_such_procedure'
                                        ]:
                                            # anything but "no_such_object" is unexpected (and fatal)
                                            raise
                                        self.log.warn(
                                            '{func} No rlink {runtime_rlink_id} currently running for router cluster worker {worker_name}: starting rlink ..',
                                            func=hltype(self._apply_routercluster_placements),
                                            worker_name=hlid(worker_name),
                                            runtime_rlink_id=hlid(runtime_rlink_id))
                                    else:
                                        self.log.info(
                                            '{func} Ok, rlink {runtime_rlink_id} already running on router cluster worker {worker_name}',
                                            func=hltype(self._apply_routercluster_placements),
                                            worker_name=hlid(worker_name),
                                            runtime_rlink_id=hlid(runtime_rlink_id))

                                    if not running_rlink:
                                        if not other_placement.tcp_listening_port or not other_node.cluster_ip:
                                            self.log.warn(
                                                '{func} Missing rlink target cluster listening port or IP in placement (cluster_ip={cluster_ip}, tcp_listening_port={tcp_listening_port})',
                                                cluster_ip=hlval(other_node.cluster_ip),
                                                tcp_listening_port=hlval(other_placement.tcp_listening_port),
                                                func=hltype(self._check_and_apply))
                                            is_running_completely = False
                                        else:
                                            rlink_config = {
                                                'realm': realm_name,
                                                'authid': node_authid,
                                                'transport': {
                                                    'type':
                                                    'rawsocket',
                                                    'endpoint': {
                                                        'type': 'tcp',
                                                        'host': other_node.cluster_ip,
                                                        'port': other_placement.tcp_listening_port
                                                    },
                                                    'serializer':
                                                    'cbor',
                                                    'url':
                                                    'rs://{}:{}'.format(other_node.cluster_ip,
                                                                        other_placement.tcp_listening_port),
                                                },
                                                'forward_local_invocations': True,
                                                'forward_remote_invocations': False,
                                                'forward_local_events': True,
                                                'forward_remote_events': False,
                                            }
                                            rlink_started = yield self._manager._session.call(
                                                'crossbarfabriccenter.remote.router.start_router_realm_link',
                                                str(other_node_oid), other_worker_name, runtime_realm_id,
                                                runtime_rlink_id, rlink_config)

                                            running_rlink = yield self._manager._session.call(
                                                'crossbarfabriccenter.remote.router.get_router_realm_link',
                                                str(other_node_oid), other_worker_name, runtime_realm_id,
                                                runtime_rlink_id)

                                            self.log.info(
                                                '{func} Rlink {runtime_rlink_id} started on router cluster worker {worker_name}:\n{rlink_started}\n{running_rlink}',
                                                func=hltype(self._apply_routercluster_placements),
                                                rlink_started=pformat(rlink_started),
                                                running_rlink=pformat(running_rlink),
                                                worker_name=hlid(worker_name),
                                                runtime_rlink_id=hlid(runtime_rlink_id))

            else:
                if node:
                    self.log.warn('{func} Router cluster node {node_oid} not running [status={status}]',
                                  func=hltype(self._apply_routercluster_placements),
                                  node_oid=hlid(node_oid),
                                  status=hl(node.status if node else 'offline'))
                else:
                    self.log.warn('{func} Router cluster node {node_oid} from placement not found! [nodes={nodes}]',
                                  func=hltype(self._apply_routercluster_placements),
                                  node_oid=node_oid,
                                  nodes=list(self._manager._session.nodes.keys()))

                # if we are missing a node we expect, we didn't run completely successfully
                is_running_completely = False

        # if we processed everything (we expected) successfully, then return True
        returnValue(is_running_completely)


class ApplicationRealmManager(object):
    """
    Application realm manager. Application realms serve as isolated WAMP routing namespaces
    together with client authentication and authorization configuration.

    Application realms can have the following entities added:

    * application roles with permissions
    * application principals with credentials

    An application realm can be started on a router worker group run on a router cluster.
    """
    log = make_logger()

    # publication options for management API events
    _PUBOPTS = PublishOptions(acknowledge=True)

    def __init__(self,
                 session,
                 globaldb: zlmdb.Database,
                 globalschema: cfxdb.globalschema.GlobalSchema,
                 db: zlmdb.Database,
                 schema: cfxdb.mrealmschema.MrealmSchema,
                 reactor=None):
        """

        :param session: Backend of user created management realms.

        :param globaldb: Global database handle.

        :param globalschema: Global database schema.

        :param db: Management realm database handle.

        :param schema: Management realm database schema.
        """
        self._session = session

        # ContainerController
        self._worker = session.config.controller

        # node personality
        self._personality = self._worker.personality

        # twisted reactor
        self._reactor = self._worker._reactor
        # if not reactor:
        #     from twisted.internet import reactor
        # self._reactor = reactor

        # global database handle & schema

        # global (node level) database handle & schema
        self.gdb = globaldb
        self.gschema = globalschema

        # mrealm database handle & schema
        self.db = db
        self.schema = schema

        # will be set in session.register
        self._prefix = None

        # the management realm OID this manager operates for
        self._mrealm_oid = session._mrealm_oid

        # filled when started
        self._started = None

        # application realm monitors, containing a map, for every application realm in state STARTING or RUNNING
        # with objects of class ApplicationRealmMonitor
        self._monitors: Dict[uuid.UUID, ApplicationRealmMonitor] = {}

    @inlineCallbacks
    def start(self, prefix):
        assert self._started is None, 'cannot start arealm manager - already running!'
        assert self._prefix is None

        self._started = time_ns()
        # crossbarfabriccenter.mrealm.arealm
        self._prefix = prefix[:-1] if prefix.endswith('.') else prefix

        regs = yield self._session.register(self,
                                            prefix='{}.'.format(self._prefix),
                                            options=RegisterOptions(details_arg='details'))
        procs = [reg.procedure for reg in regs]
        self.log.info(
            'Mrealm controller {api} registered management procedures prefix "{prefix}" [{func}]:\n\n{procs}\n',
            api=hl('Application realm manager API', color='green', bold=True),
            func=hltype(self.start),
            prefix=hlval(self._prefix),
            procs=hl(pformat(procs), color='white', bold=True))

        # start all application realm monitors ..
        cnt_started = 0
        cnt_skipped = 0
        dl = []
        with self.db.begin() as txn:
            arealm_oids = self.schema.arealms.select(txn, return_values=False)
            for arealm_oid in arealm_oids:
                arealm = self.schema.arealms[txn, arealm_oid]
                if arealm.status in [ApplicationRealm.STATUS_STARTING, ApplicationRealm.STATUS_RUNNING]:
                    assert arealm_oid not in self._monitors
                    monitor = ApplicationRealmMonitor(self, arealm_oid)
                    dl.append(monitor.start())
                    self._monitors[arealm_oid] = monitor
                    cnt_started += 1
                    self.log.info(
                        '{func}(prefix="{prefix}"): {action} for application realm {arealm_oid} in {status})',
                        action=hl('cluster monitor started', color='green', bold=True),
                        prefix=hlval(prefix),
                        func=hltype(self.start),
                        arealm_oid=hlid(arealm_oid),
                        status=hlval(arealm.status))
                else:
                    cnt_skipped += 1
                    self.log.info(
                        '{func}(prefix="{prefix}"): {action} for application realm {arealm_oid} in status {status}',
                        action=hl('cluster monitor skipped', color='green', bold=True),
                        prefix=hlval(prefix),
                        func=hltype(self.start),
                        arealm_oid=hlid(arealm_oid),
                        status=hlval(arealm.status))
        self.log.info(
            'Application realm manager has started monitors for {cnt_started} application realms ({cnt_skipped} skipped) [{func}]',
            cnt_started=hlval(cnt_started),
            cnt_skipped=hlval(cnt_skipped),
            func=hltype(self.start))

        self.log.info('Application realm manager ready for management realm {mrealm_oid}! [{func}]',
                      mrealm_oid=hlid(self._mrealm_oid),
                      func=hltype(self.start))

        # return txaio.gather(dl)

    @inlineCallbacks
    def stop(self):
        assert self._started > 0, 'cannot stop arealm manager - currently not running!'

        # stop all application realm monitors ..
        dl = []
        for arealm_oid, arealm_monitor in self._monitors.items():
            dl.append(arealm_monitor.stop())
            del self._monitors[arealm_oid]
        self._started = None
        self.log.info(
            'Ok, application realm manager for management realm {mrealm_oid} stopped ({cnt_stopped} monitors stopped) [{func}]',
            mrealm_oid=hlid(self._mrealm_oid),
            cnt_stopped=len(dl),
            func=hltype(self.start))

        # return txaio.gather(dl)

    @wamp.register(None, check_types=True)
    def list_arealms(self, return_names: Optional[bool] = None, details: Optional[CallDetails] = None) -> List[str]:
        """
        Returns list of application realms defined. Detail information for an application realm
        can be retrieved using :meth:`crossbar.master.arealm.ApplicationRealmManager.get_arealm`.

        :param return_names: Return application realm names instead of  object IDs

        :return: List of application realm object IDs or names. For example:

            .. code-block:: json

                ["788cda65-a41d-49ae-b09a-b51967a34915",
                 "4b051f7a-1733-4784-aade-35dbdab6a234",
                 "0bb832f4-9fd0-4916-8ef1-a5799564f5fc"]
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(details={details})', func=hltype(self.list_arealms), details=details)

        with self.db.begin() as txn:
            if return_names:
                arealms = self.schema.arealms.select(txn, return_keys=False)
                if arealms:
                    return sorted([arealm.name for arealm in arealms])
                else:
                    return []
            else:
                arealm_oids = self.schema.arealms.select(txn, return_values=False)
                if arealm_oids:
                    # we now have a list of uuid.UUID objects: convert to strings
                    return [str(oid) for oid in arealm_oids]
                else:
                    return []

    @wamp.register(None, check_types=True)
    def get_arealm(self, arealm_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Return configuration and run-time status information for an application realm (by object ID).

        :param arealm_oid: Object ID of the application realm to return.

        :return: Application realm definition. For example, initially, after an application
            realm has been created:

            .. code-block:: json

                {"bridge_meta_api": true,
                 "changed": null,
                 "description": null,
                 "enable_meta_api": true,
                 "label": null,
                 "name": "myrealm1",
                 "oid": "f1b58365-f936-4c4f-a820-9c5ed06eacfb",
                 "owner": null,
                 "status": "STOPPED",
                 "tags": null,
                 "webcluster_oid": null,
                 "workergroup_oid": null}

            Once the application realm has been started on a router worker group and web cluster,
            the ``status``, ``webcluster_oid`` and ``workergroup_oid`` attributes will change
            accordingly:

            .. code-block:: json

                {"bridge_meta_api": true,
                 "changed": null,
                 "description": null,
                 "enable_meta_api": true,
                 "label": null,
                 "name": "myrealm1",
                 "oid": "f1b58365-f936-4c4f-a820-9c5ed06eacfb",
                 "owner": null,
                 "status": "RUNNING",
                 "tags": null,
                 "webcluster_oid": "2f279b0f-e65a-4d7d-bef2-983f3b723e95",
                 "workergroup_oid": "74ba0f88-eb7d-4810-a901-4a6d611d7519"}

        """
        assert type(arealm_oid) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(arealm_oid={arealm_oid}, details={details})',
                      func=hltype(self.get_arealm),
                      arealm_oid=hlid(arealm_oid),
                      details=details)

        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid arealm_oid: {}'.format(str(e)))

        with self.db.begin() as txn:
            arealm = self.schema.arealms[txn, arealm_oid_]

        if arealm:
            return arealm.marshal()
        else:
            raise ApplicationError('crossbar.error.no_such_object', 'no arealm with oid {}'.format(arealm_oid_))

    @wamp.register(None, check_types=True)
    def get_arealm_by_name(self, arealm_name: str, details: Optional[CallDetails] = None):
        """
        Return configuration and run-time status information for an application realm (by name).

        See also the corresponding procedure :meth:`crossbar.master.arealm.ApplicationRealmManager.get_arealm`
        which returns the same information, given and object ID rather than name.

        :param arealm_name: Name (WAMP realm) of the application realm to return.

        :return: Application realm definition.
        """
        assert type(arealm_name) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(arealm_name="{arealm_name}", details={details})',
                      func=hltype(self.get_arealm_by_name),
                      arealm_name=hlid(arealm_name),
                      details=details)

        with self.db.begin() as txn:
            arealm_oid = self.schema.idx_arealms_by_name[txn, arealm_name]
            if not arealm_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm named {}'.format(arealm_name))

            arealm = self.schema.arealms[txn, arealm_oid]
            assert arealm

        return arealm.marshal()

    @wamp.register(None, check_types=True)
    async def create_arealm(self, arealm: dict, details: Optional[CallDetails] = None) -> dict:
        """
        Create a new application realm definition.

        :procedure: ``crossbarfabriccenter.mrealm.arealm.create_arealm`` URI of WAMP procedure to call.
        :event: ``crossbarfabriccenter.mrealm.arealm.on_arealm_created`` WAMP event published once the
            application realm has been created.
        :error: ``wamp.error.invalid_configuration`` WAMP error returned when the application realm
            configuration provided has a problem.
        :error: ``wamp.error.not_authorized`` WAMP error returned when the user is currently not allowed
            to created (another) application realm.

        :param arealm: Application realm definition. For example:

            .. code-block:: json

                {
                    "name": "myrealm1",
                    "enable_meta_api": true,
                    "bridge_meta_api": true
                }

        :return: Application realm creation information.
        """
        assert type(arealm) == dict
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(arealm="{arealm}", details={details})',
                      func=hltype(self.create_arealm),
                      arealm=arealm,
                      details=details)

        obj = ApplicationRealm.parse(arealm)

        # object ID of new application realm
        obj.oid = uuid.uuid4()

        if details and details.caller_authid:
            with self.gdb.begin() as txn:
                caller_oid = self.gschema.idx_users_by_email[txn, details.caller_authid]
                if not caller_oid:
                    raise ApplicationError('wamp.error.no_such_principal',
                                           'no user found for authid "{}"'.format(details.caller_authid))
            obj.owner_oid = caller_oid
        else:
            raise ApplicationError('wamp.error.no_such_principal', 'cannot map user - no caller authid available')

        # unless and until the application realm is started, no router worker
        # group or web cluster is assigned
        obj.workergroup_oid = None
        obj.webcluster_oid = None

        # if this arealm is federated, it is associated with a specific XBR datamarket
        if obj.datamarket_oid:
            # FIXME: check for datamarket_oid ..
            self.log.info('new application realm is associated datamarket_oid {datamarket_oid}',
                          datamarket_oid=hlid(obj.datamarket_oid))

        # set initial status of application realm
        obj.status = ApplicationRealm.STATUS_STOPPED
        obj.changed = np.datetime64(time_ns(), 'ns')

        with self.db.begin(write=True) as txn:
            self.schema.arealms[txn, obj.oid] = obj

        self.log.info('new application realm object stored in database:\n{obj}', obj=obj)

        res_obj = obj.marshal()

        await self._session.publish('{}.on_arealm_created'.format(self._prefix), res_obj, options=self._PUBOPTS)

        self.log.info('Management API event <on_arealm_created> published:\n{res_obj}', res_obj=res_obj)

        return res_obj

    @wamp.register(None, check_types=True)
    async def delete_arealm(self, arealm_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Delete an existing application realm definition.

        :procedure: ``crossbarfabriccenter.mrealm.arealm.delete_arealm`` URI of WAMP procedure to call.
        :event: ``crossbarfabriccenter.mrealm.arealm.on_arealm_deleted`` WAMP event published once the application realm has been deleted.
        :error: ``wamp.error.invalid_argument`` WAMP error returned when ``arealm_oid`` was invalid.
        :error: ``crossbar.error.no_such_object`` WAMP error returned when ``arealm_oid`` was not found.
        :error: ``crossbar.error.not_stopped`` WAMP error returned when application realm is not in status ``STOPPED``.

        :param arealm_oid: OID of the application realm to delete

        :returns: Application realm deletion information.
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(arealm_oid={arealm_oid}, details={details})',
                      func=hltype(self.delete_arealm),
                      arealm_oid=hlid(arealm_oid),
                      details=details)

        try:
            oid = uuid.UUID(arealm_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        with self.db.begin(write=True) as txn:
            arealm = self.schema.arealms[txn, oid]
            if arealm:
                if arealm.status != ApplicationRealm.STATUS_STOPPED:
                    raise ApplicationError(
                        'crossbar.error.not_stopped',
                        'application realm with oid {} found, but currently in status "{}"'.format(oid, arealm.status))
                del self.schema.arealms[txn, oid]
            else:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(oid))

        self.log.info('Application realm object deleted from database:\n{arealm}', arealm=arealm)

        arealm_obj = arealm.marshal()

        await self._session.publish('{}.on_arealm_deleted'.format(self._prefix), arealm_obj, options=self._PUBOPTS)

        return arealm_obj

    @wamp.register(None, check_types=True)
    async def start_arealm(self,
                           arealm_oid: str,
                           router_workergroup_oid: str,
                           webcluster_oid: str,
                           details: Optional[CallDetails] = None) -> dict:
        """
        Start an application realm on a router cluster worker group and webcluster.

        The webcluster is responsible for accepting frontend client connections, performing
        WAMP authentication, selecting a backend router worker (from the router worker group) and
        forward application traffic of the connected frontend session to the backend router worker.

        Clients will be able to authenticate (to the frontend webcluster) using any credentials
        defined on the application realm. When a client successfully authenticates using one of the
        credentials defined, it will be identified as the principal on the application realm
        associated with the credential.

        A principal in turn will have an associated role defined, and the permissions on that role
        ultimately then determine the rights granted to the client to perform respective WAMP actions
        (eg call "com.example.add2" or subscribe "com.example.onevent1") on URIs within the
        application realm.

        The application realm monitor will take care of:

        1. start the application realm on all router workers on nodes of the respective router worker group
        2. start all roles defined for the application realm on those router workers
        3. start router-to-router links between the router workers in the router worker group for the respective application realm
        4. start backend connections and routes to the router workers from the proxy workers of the webcluster
        5. configure credentials for the principals on the application realm in the proxy workers of the webcluster

        :param arealm_oid: The application realm to start.
        :param router_workergroup_oid: The router cluster worker group to start the application realm on.
        :param webcluster_oid: The web cluster to serve as a frontend layer for the application realm.

        :return: Application realm start information.
        """
        self.log.info(
            '{func}(arealm_oid="{arealm_oid}", router_workergroup_oid="{router_workergroup_oid}", details={details})',
            func=hltype(self.start_arealm),
            arealm_oid=hlid(arealm_oid),
            router_workergroup_oid=hlid(router_workergroup_oid),
            details=details)

        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid arealm_oid "{}"'.format(arealm_oid))

        try:
            router_workergroup_oid_ = uuid.UUID(router_workergroup_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument',
                                   'invalid workergroup_oid "{}"'.format(router_workergroup_oid))

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid webcluster_oid "{}"'.format(webcluster_oid))

        with self.db.begin(write=True) as txn:
            arealm = self.schema.arealms[txn, arealm_oid_]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(arealm_oid_))

            router_workergroup = self.schema.router_workergroups[txn, router_workergroup_oid_]
            if not router_workergroup:
                raise ApplicationError(
                    'crossbar.error.no_such_object',
                    'no router cluster worker group with oid {} found'.format(router_workergroup_oid_))

            webcluster = self.schema.webclusters[txn, webcluster_oid_]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no webcluster with oid {} found'.format(webcluster_oid_))

            if arealm.status not in [cfxdb.mrealm.types.STATUS_STOPPED, cfxdb.mrealm.types.STATUS_PAUSED]:
                emsg = 'cannot start arealm currently in state {}'.format(
                    cfxdb.mrealm.types.STATUS_BY_CODE[arealm.status])
                raise ApplicationError('crossbar.error.cannot_start', emsg)

            arealm.status = cfxdb.mrealm.types.STATUS_STARTING
            arealm.workergroup_oid = router_workergroup_oid_
            arealm.webcluster_oid = webcluster_oid_
            arealm.changed = np.datetime64(time_ns(), 'ns')

            self.schema.arealms[txn, arealm_oid_] = arealm

        assert arealm_oid not in self._monitors
        monitor = ApplicationRealmMonitor(self, arealm_oid_)
        monitor.start()
        self._monitors[arealm_oid_] = monitor

        arealm_starting = {
            'oid': str(arealm.oid),
            'status': cfxdb.mrealm.types.STATUS_BY_CODE[arealm.status],
            'router_workergroup_oid': str(router_workergroup_oid_),
            'changed': int(arealm.changed) if arealm.changed else None,
            'who': {
                'session': details.caller if details else None,
                'authid': details.caller_authid if details else None,
                'authrole': details.caller_authrole if details else None,
            }
        }
        await self._session.publish('{}.on_arealm_starting'.format(self._prefix),
                                    arealm_starting,
                                    options=self._PUBOPTS)

        return arealm_starting

    @wamp.register(None, check_types=True)
    async def stop_arealm(self, arealm_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Stop a currently running application realm. This will stop router workers started
        in the worker group assigned for the application realm, and remove the association
        with the proxy workers of the web cluster responsible for the application realm.

        :event: ``crossbarfabriccenter.mrealm.arealm.on_arealm_stopping`` WAMP event published once the application realm is stopping.
        :error: ``wamp.error.invalid_argument`` WAMP error returned when ``arealm_oid`` was invalid.
        :error: ``crossbar.error.no_such_object`` WAMP error returned when ``arealm_oid`` was not found.
        :error: ``crossbar.error.cannot_stop`` WAMP error returned when application realm cannot be stopped,
            because it is not in status ``RUNNING`` or ``STARTING``.

        :param arealm_oid: The application realm to stop.

        :return: Application realm stopping information.
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(arealm_oid={arealm_oid}, details={details})',
                      arealm_oid=hlid(arealm_oid),
                      func=hltype(self.stop_arealm),
                      details=details)

        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        with self.db.begin(write=True) as txn:
            arealm = self.schema.arealms[txn, arealm_oid_]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(arealm_oid_))

            if arealm.status not in [cfxdb.mrealm.types.STATUS_STARTING, cfxdb.mrealm.types.STATUS_RUNNING]:
                emsg = 'cannot stop arealm currently in state {}'.format(
                    cfxdb.mrealm.types.STATUS_BY_CODE[arealm.status])
                raise ApplicationError('crossbar.error.cannot_stop', emsg)

            arealm.status = cfxdb.mrealm.types.STATUS_STOPPING
            arealm.changed = np.datetime64(time_ns(), 'ns')

            self.schema.arealms[txn, arealm_oid_] = arealm

        arealm_stopping = {
            'oid': str(arealm.oid),
            'status': cfxdb.mrealm.types.STATUS_BY_CODE[arealm.status],
            'changed': arealm.changed,
            'who': {
                'session': details.caller if details else None,
                'authid': details.caller_authid if details else None,
                'authrole': details.caller_authrole if details else None,
            }
        }

        await self._session.publish('{}.on_arealm_stopping'.format(self._prefix),
                                    arealm_stopping,
                                    options=self._PUBOPTS)

        return arealm_stopping

    @wamp.register(None, check_types=True)
    def stat_arealm(self, arealm_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        *NOT YET IMPLEMENTED*

        Get current status and statistics for given application realm.

        :param arealm_oid: The application realm to return status and statistics for.

        :return: Current status and statistics for given routercluster.
        """
        assert type(arealm_oid) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(arealm_oid={arealm_oid}, details={details})',
                      func=hltype(self.stat_arealm),
                      arealm_oid=hlid(arealm_oid),
                      details=details)

        raise NotImplementedError()

    @wamp.register(None, check_types=True)
    def list_router_workers(self, arealm_oid: str, details: Optional[CallDetails] = None) -> List[str]:
        """
        *NOT YET IMPLEMENTED*

        When an application realm has been started on a router workergroup and is running,
        the list of router workers (in the router worker group) the application realm
        is hosted on.

        :param arealm_oid: Object ID of application realm to list router workers for.

        :return: List of router workers in the router worker group running the application realm.
        """
        raise NotImplementedError()

    @wamp.register(None, check_types=True)
    def list_principals(self,
                        arealm_oid: str,
                        return_names: Optional[bool] = None,
                        details: Optional[CallDetails] = None) -> List[str]:
        """
        List all principals defined on this application realm.

        :param arealm_oid: Object ID of application realm to list principals for.
        :param return_names: Return principal names (WAMP authids) rather than object IDs.

        :return: List of principal object IDs or names.
        """
        assert type(arealm_oid) == str
        assert return_names is None or type(return_names) == bool
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(arealm_oid={arealm_oid}, details={details})',
                      func=hltype(self.list_principals),
                      arealm_oid=hlid(arealm_oid),
                      details=details)

        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(arealm_oid))

        with self.db.begin() as txn:
            arealm = self.schema.arealms[txn, arealm_oid_]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(arealm_oid_))

            if return_names:
                res = []
                for _, principal_name in self.schema.idx_principals_by_name.select(
                        txn,
                        from_key=(arealm_oid_, ''),
                        to_key=(uuid.UUID(int=(int(arealm_oid_) + 1)), ''),
                        return_values=False):
                    res.append(principal_name)
                return sorted(res)
            else:
                res = []
                for principal_oid in self.schema.idx_principals_by_name.select(
                        txn,
                        from_key=(arealm_oid_, ''),
                        to_key=(uuid.UUID(int=(int(arealm_oid_) + 1)), ''),
                        return_keys=False):
                    res.append(str(principal_oid))
                return res

    @wamp.register(None, check_types=True)
    def get_principal(self, arealm_oid: str, principal_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Return definition of principal.

        :param arealm_oid: Object ID of application realm the principal is defined on.
        :param principal_oid: Object ID of the principal to return.

        :return: Principal definition.
        """
        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(arealm_oid))

        try:
            principal_oid_ = uuid.UUID(principal_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(principal_oid))

        with self.db.begin() as txn:
            arealm = self.schema.arealms[txn, arealm_oid_]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(arealm_oid_))

            principal = self.schema.principals[txn, principal_oid_]
            if not principal or principal.arealm_oid != arealm_oid_:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no principal with oid {} found in application realm'.format(principal_oid_))

        obj = principal.marshal()
        return obj

    @wamp.register(None, check_types=True)
    def get_principal_by_name(self, arealm_oid: str, principal_name: str, details: Optional[CallDetails] = None):
        """
        Return definition of principal by principal name (WAMP authid).

        :param arealm_oid: Object ID of application realm the principal is defined on.
        :param principal_oid: Object ID of the principal to return.

        :return: Principal definition.
        """
        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(arealm_oid))

        with self.db.begin() as txn:
            arealm = self.schema.arealms[txn, arealm_oid_]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(arealm_oid_))

            principal_oid = self.schema.idx_principals_by_name[txn, (arealm_oid_, principal_name)]
            if not principal_oid:
                raise ApplicationError(
                    'crossbar.error.no_such_object',
                    'no principal with name (authid) "{}" found in application realm'.format(principal_name))

            principal = self.schema.principals[txn, principal_oid]
            if not principal or principal.arealm_oid != arealm_oid_:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no principal with oid {} found in application realm'.format(principal_oid))

        obj = principal.marshal()
        return obj

    @wamp.register(None, check_types=True)
    async def add_principal(self, arealm_oid: str, principal: dict, details: Optional[CallDetails] = None) -> dict:
        """
        Add a new principal to the given application realm.

        :param arealm_oid: Object ID of application realm to add the principal to.
        :param principal: Principal definition.

        :return: Principal addition information.
        """
        assert type(principal) == dict
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(arealm_oid="{arealm_oid}", principal={principal}, details={details})',
                      func=hltype(self.add_principal),
                      arealm_oid=hlid(arealm_oid),
                      principal=principal,
                      details=details)

        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(arealm_oid))

        role_oid = principal.get('role_oid', None)
        assert role_oid is not None and type(role_oid) == str
        try:
            role_oid_ = uuid.UUID(role_oid)
        except:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(role_oid))

        try:
            obj = Principal.parse(principal)
        except Exception as e:
            raise ApplicationError('crossbar.error.invalid_config', 'invalid configuration. {}'.format(e))

        obj.oid = uuid.uuid4()
        obj.modified = np.datetime64(time_ns(), 'ns')
        obj.arealm_oid = arealm_oid_
        obj.role_oid = role_oid_

        with self.db.begin(write=True) as txn:
            arealm = self.schema.arealms[txn, obj.arealm_oid]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(obj.arealm_oid))

            role = self.schema.roles[txn, obj.role_oid]
            if not role:
                raise ApplicationError('crossbar.error.no_such_object', 'no role with oid {} found'.format(role_oid_))
            principal_oid = self.schema.idx_principals_by_name[txn, (obj.arealm_oid, obj.authid)]
            if principal_oid:
                raise ApplicationError(
                    'crossbar.error.already_exists',
                    'principal with name (authid) "{}" already exist in application realm'.format(obj.authid))

            self.schema.principals[txn, obj.oid] = obj

        self.log.info('new principal object stored in database:\n{obj}', obj=obj)

        res_obj = obj.marshal()

        await self._session.publish('{}.on_principal_created'.format(self._prefix), res_obj, options=self._PUBOPTS)

        self.log.info('Management API event <on_principal_created> published:\n{res_obj}', res_obj=res_obj)

        return res_obj

    @wamp.register(None, check_types=True)
    def remove_principal(self, arealm_oid: str, principal_oid: str, details: Optional[CallDetails] = None):
        """
        Remove a principal from the given application realm.

        :param arealm_oid: Object ID of application realm from which to remove the principal.
        :param principal_oid: Object ID of the principal to remove.

        :return: Principal removal information.
        """
        raise NotImplementedError()

    @wamp.register(None, check_types=True)
    def list_principal_credentials(self,
                                   arealm_oid: str,
                                   principal_oid: str,
                                   details: Optional[CallDetails] = None) -> List[str]:
        """
        List credentials for a principal.

        :param arealm_oid: Object ID of application realm of the principal to list credentials for.
        :param principal_oid: Object ID of the principal to list credentials for.

        :return: List of credential object IDs or names (WAMP authids).
        """
        assert type(arealm_oid) == str
        assert type(principal_oid) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(arealm_oid={arealm_oid}, principal_oid={principal_oid}, details={details})',
                      func=hltype(self.list_principal_credentials),
                      arealm_oid=hlid(arealm_oid),
                      principal_oid=hlid(principal_oid),
                      details=details)

        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid arealm_oid "{}"'.format(arealm_oid))

        try:
            principal_oid_ = uuid.UUID(principal_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid principal_oid "{}"'.format(principal_oid))

        with self.db.begin() as txn:
            arealm = self.schema.arealms[txn, arealm_oid_]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(arealm_oid_))

            principal = self.schema.principals[txn, principal_oid_]
            if not principal or principal.arealm_oid != arealm_oid_:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no principal with oid {} found in application realm'.format(principal_oid_))

            res = []
            for credential_oid in self.schema.idx_credentials_by_principal.select(
                    txn,
                    from_key=(principal_oid_, np.datetime64(0, 'ns')),
                    to_key=(uuid.UUID(int=(int(principal_oid_) + 1)), np.datetime64(0, 'ns')),
                    return_keys=False):
                res.append(str(credential_oid))
            return res

    @wamp.register(None, check_types=True)
    def get_principal_credential(self,
                                 arealm_oid: str,
                                 principal_oid: str,
                                 credential_oid: str,
                                 details: Optional[CallDetails] = None) -> dict:
        """
        Return definition of a credential of a principal.

        :param arealm_oid: Object ID of application realm of the principal to return a credential for.
        :param principal_oid: Object ID of the principal to get a credentials for.
        :param credential_oid: Object ID of the credential to return.

        :return: Credential detail information.
        """
        raise NotImplementedError()

    @wamp.register(None, check_types=True)
    def add_principal_credential(self,
                                 arealm_oid: str,
                                 principal_oid: str,
                                 credential: dict,
                                 details: Optional[CallDetails] = None) -> dict:
        """
        Add credentials to a principal.

        :param arealm_oid: Object ID of application realm of the principal to add a credential for.
        :param principal_oid: Object ID of the principal to add a credentials for.
        :param credential: Credential configuration. Examples:

            **WAMP-anonymous:**

            .. code-block:: json

                {
                    "authmethod": "anonymous"
                }

            **WAMP-ticket:**

            .. code-block:: json

                {
                    "authmethod": "ticket",
                    "secret": "secret123"
                }

            **WAMP-wampcra:**

            .. code-block:: json

                {
                    "authmethod": "wampcra",
                    "secret": "secret123",
                    "salt": "salt456",
                    "iterations": 100,
                    "keylen": 16
                }

            **WAMP-scram:**

            .. code-block:: json

                {
                    "authmethod": "scram",
                    "kdf": "pbkdf2",
                    "iterations": 100,
                    "memory": 0,
                    "stored-key": "",
                    "server-key": ""
                }

            **WAMP-cryptosign:**

            .. code-block:: json

                {
                    "authmethod": "cryptosign",
                    "authconfig": {
                        "authorized_keys": [
                            "92b450bb5fb168b396ad2bde633825662665b4cb73c1243ce5e834971c9354f5"
                        ]
                    }
                }

            **WAMP-tls:** *NOT YET IMPLEMENTED*

            .. code-block:: json

                {
                    "authmethod": "tls"
                }

            **WAMP-cookie:** *NOT YET IMPLEMENTED*

            .. code-block:: json

                {
                    "authmethod": "cookie"
                }

        :return: Credential addition information.
        """
        self.log.info(
            '{func}(arealm_oid="{arealm_oid}", principal_oid="{principal_oid}", credential={credential}, details={details})',
            func=hltype(self.add_principal_credential),
            arealm_oid=hlid(arealm_oid),
            principal_oid=hlid(principal_oid),
            credential=pformat(credential),
            details=details)

        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid arealm_oid "{}"'.format(arealm_oid))

        try:
            principal_oid_ = uuid.UUID(principal_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid principal_oid "{}"'.format(principal_oid))

        with self.db.begin(write=True) as txn:
            arealm = self.schema.arealms[txn, arealm_oid_]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(arealm_oid_))

            principal = self.schema.principals[txn, principal_oid_]
            if not principal or principal.arealm_oid != arealm_oid_:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no principal with oid {} found in application realm'.format(principal_oid_))

            try:
                credential_ = Credential.parse(credential)
            except Exception as e:
                raise ApplicationError('crossbar.error.invalid_config', 'invalid configuration. {}'.format(e))

            credential_.oid = uuid.uuid4()
            credential_.created = np.datetime64(time_ns(), 'ns')
            credential_.realm = arealm.name
            credential_.authid = principal.authid
            credential_.principal_oid = principal_oid_

            if self.schema.idx_credentials_by_auth[txn,
                                                   (credential_.authmethod, credential_.realm, credential_.authid)]:
                raise ApplicationError(
                    'crossbar.error.already_running',
                    'duplicate credential for authmethod "{}", realm "{}" and authid "{}" in application realm'.format(
                        credential_.authmethod, credential_.realm, credential_.authid))

            self.schema.credentials[txn, credential_.oid] = credential_

        return credential_.marshal()

    @wamp.register(None, check_types=True)
    def remove_principal_credential(self,
                                    arealm_oid: str,
                                    principal_oid: str,
                                    credential_oid: str,
                                    details: Optional[CallDetails] = None) -> dict:
        """
        Remove credentials from a principal.

        :param arealm_oid: Object ID of application realm of the principal to remove a credential from.
        :param principal_oid: Object ID of the principal to remove a credentials from.
        :param credential_oid: Object ID of the credential to remove.

        :return: Credential removal information.
        """
        raise NotImplementedError()

    @wamp.register(None, check_types=True)
    def list_roles(self, return_names: Optional[bool] = None, details: Optional[CallDetails] = None) -> List[str]:
        """
        Returns list of roles defined.

        :param return_names: Return roles names instead of  object IDs

        :return: List of role object IDs or names.
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(return_names={return_names}, details={details})',
                      func=hltype(self.list_roles),
                      return_names=hlval(return_names),
                      details=details)

        with self.db.begin() as txn:
            if return_names:
                roles = self.schema.roles.select(txn, return_keys=False)
                if roles:
                    return sorted([role.name for role in roles])
                else:
                    return []
            else:
                role_oids = self.schema.roles.select(txn, return_values=False)
                if role_oids:
                    # we now have a list of uuid.UUID objects: convert to strings
                    return [str(oid) for oid in role_oids]
                else:
                    return []

    @wamp.register(None, check_types=True)
    def get_role(self, role_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Return configuration information for a role.

        :param role_oid: Object ID of the role to return.

        :return: Role definition.
        """
        assert type(role_oid) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(role_oid={role_oid}, details={details})',
                      func=hltype(self.get_role),
                      role_oid=hlid(role_oid),
                      details=details)

        try:
            role_oid_ = uuid.UUID(role_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid role_oid: {}'.format(str(e)))

        with self.db.begin() as txn:
            role = self.schema.roles[txn, role_oid_]

        if role:
            return role.marshal()
        else:
            raise ApplicationError('crossbar.error.no_such_object', 'no role with oid {}'.format(role_oid_))

    @wamp.register(None, check_types=True)
    def get_role_by_name(self, role_name: str, details: Optional[CallDetails] = None):
        """
        Return configuration information for a role given by name.

        :param role_name: The name of the role to return the definition for.

        :return: Role definition.
        """
        assert type(role_name) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(role_name="{role_name}", details={details})',
                      func=hltype(self.get_role_by_name),
                      role_name=hlid(role_name),
                      details=details)

        with self.db.begin() as txn:
            role_oid = self.schema.idx_roles_by_name[txn, role_name]
            if not role_oid:
                raise ApplicationError('crossbar.error.no_such_object', 'no role named "{}"'.format(role_name))

            role = self.schema.roles[txn, role_oid]
            assert role

        return role.marshal()

    @wamp.register(None, check_types=True)
    async def create_role(self, role: dict, details: Optional[CallDetails] = None) -> dict:
        """
        Create a new Role definition.

        :param role: Role definition.

        :return: Role creation information.
        """
        assert type(role) == dict
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(role="{role}", details={details})',
                      func=hltype(self.create_role),
                      role=role,
                      details=details)

        obj = Role.parse(role)
        obj.oid = uuid.uuid4()
        obj.created = np.datetime64(time_ns(), 'ns')

        with self.db.begin(write=True) as txn:
            self.schema.roles[txn, obj.oid] = obj

        self.log.info('new role object stored in database:\n{obj}', obj=obj)

        res_obj = obj.marshal()

        await self._session.publish('{}.on_role_created'.format(self._prefix), res_obj, options=self._PUBOPTS)

        self.log.info('Management API event <on_role_created> published:\n{res_obj}', res_obj=res_obj)

        return res_obj

    @wamp.register(None, check_types=True)
    async def delete_role(self, role_oid: str, details: Optional[CallDetails] = None):
        """
        Delete an existing Role definition.

        :procedure: ``crossbarfabriccenter.mrealm.arealm.delete_role``
        :event: ``crossbarfabriccenter.mrealm.arealm.on_role_deleted``
        :error: ``wamp.error.invalid_argument``
        :error: ``crossbar.error.no_such_object``

        :param role_oid: OID of the Role to delete

        :returns: Role deletin information.
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(role_oid={role_oid}, details={details})',
                      func=hltype(self.delete_role),
                      role_oid=hlid(role_oid),
                      details=details)

        try:
            oid = uuid.UUID(role_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        with self.db.begin(write=True) as txn:
            obj = self.schema.roles[txn, oid]
            if obj:
                del self.schema.roles[txn, oid]
            else:
                raise ApplicationError('crossbar.error.no_such_object', 'no object with oid {} found'.format(oid))

        self.log.info('Role object deleted from database:\n{obj}', obj=obj)

        res_obj = obj.marshal()

        await self._session.publish('{}.on_role_deleted'.format(self._prefix), res_obj, options=self._PUBOPTS)

        return res_obj

    @wamp.register(None, check_types=True)
    def list_role_permissions(self,
                              role_oid: str,
                              prefix: Optional[str] = None,
                              details: Optional[CallDetails] = None) -> List[str]:
        """
        List permissions in a role.

        :param role_oid: The role to get permissions for.
        :param prefix: WAMP URI prefix of permission to filter for.

        :return: List of permissions object IDs of this role.
        """
        self.log.info('{func}(role_oid={role_oid}, prefix="{prefix}", details={details})',
                      func=hltype(self.list_role_permissions),
                      role_oid=hlid(role_oid),
                      prefix=hlval(prefix),
                      details=details)

        try:
            role_oid_ = uuid.UUID(role_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid role_oid "{}"'.format(role_oid_))

        with self.db.begin() as txn:
            role = self.schema.roles[txn, role_oid_]
            if not role:
                raise ApplicationError('crossbar.error.no_such_object', 'no role with oid {} found'.format(role_oid_))

            res = []
            from_key = (role_oid_, prefix if prefix else '')
            to_key = (uuid.UUID(int=(int(role_oid_) + 1)), '')
            for permission_oid in self.schema.idx_permissions_by_uri.select(txn,
                                                                            from_key=from_key,
                                                                            to_key=to_key,
                                                                            return_keys=False):
                res.append(str(permission_oid))

        return res

    @wamp.register(None, check_types=True)
    async def add_role_permission(self,
                                  role_oid: str,
                                  uri: str,
                                  permission: dict,
                                  details: Optional[CallDetails] = None) -> dict:
        """
        Add a permission to a role.

        :param arealm_oid: OID of the application realm to which to add the role permission.
        :param uri: WAMP URI (pattern) of the permission to add.
        :param permission: Permission definition

        :return: Permission addition information.
        """
        assert details is None or isinstance(details, CallDetails)

        try:
            role_oid_ = uuid.UUID(role_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid role_oid "{}"'.format(role_oid))

        self.log.info('{func}(role_oid={role_oid}, uri="{uri}", permission={permission}, details={details})',
                      func=hltype(self.add_role_permission),
                      role_oid=hlid(role_oid_),
                      uri=hlval(uri),
                      permission=permission,
                      details=details)

        role_permission = Permission.parse(permission)
        role_permission.oid = uuid.uuid4()
        role_permission.role_oid = role_oid_
        role_permission.uri = uri

        with self.db.begin(write=True) as txn:
            role = self.schema.roles[txn, role_permission.role_oid]
            if not role:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(role_permission.role_oid))

            self.schema.permissions[txn, role_permission.oid] = role_permission

        res_obj = role_permission.marshal()
        self.log.info('role added to permission:\n{permission}', permission=res_obj)

        await self._session.publish('{}.on_permission_added'.format(self._prefix), res_obj, options=self._PUBOPTS)

        return res_obj

    @wamp.register(None, check_types=True)
    async def remove_role_permission(self,
                                     role_oid: str,
                                     permission_oid: str,
                                     details: Optional[CallDetails] = None) -> dict:
        """
        Remove a permission from a role.

        :param role_oid: Object ID of the role from which to remove the permission.
        :param permission_oid: Object ID of the permission to remove.

        :return: Permission removal information.
        """
        assert type(role_oid) == str
        assert type(permission_oid) == str
        assert details is None or isinstance(details, CallDetails)

        try:
            role_oid_ = uuid.UUID(role_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid role_oid "{}"'.format(role_oid))

        try:
            permission_oid_ = uuid.UUID(permission_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid permission_oid "{}"'.format(permission_oid))

        self.log.info('{func}(role_oid={role_oid}, permission_oid={permission_oid}, details={details})',
                      role_oid=hlid(role_oid_),
                      permission_oid=hlid(permission_oid_),
                      func=hltype(self.remove_role_permission),
                      details=details)

        with self.db.begin(write=True) as txn:
            role = self.schema.roles[txn, role_oid_]
            if not role:
                raise ApplicationError('crossbar.error.no_such_object', 'no role with oid {} found'.format(role_oid_))

            role_permission = self.schema.permissions[txn, permission_oid_]
            del self.schema.permissions[txn, permission_oid_]

        res_obj = role_permission.marshal()
        self.log.info('role removed from arealm:\n{res_obj}', membership=res_obj)

        await self._session.publish('{}.on_arealm_role_removed'.format(self._prefix), res_obj, options=self._PUBOPTS)

        return res_obj

    @wamp.register(None, check_types=True)
    def get_role_permission(self, role_oid: str, permission_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Get information for the permission on a role.

        :param role_oid: Object ID of the role to retrieve the permission for.
        :param permission_oid: Object ID of the permission to retrieve.

        :return: Permission definition.
        """
        assert type(role_oid) == str
        assert type(permission_oid) == str
        assert details is None or isinstance(details, CallDetails)

        try:
            role_oid_ = uuid.UUID(role_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid role_oid "{}"'.format(role_oid))

        try:
            permission_oid_ = uuid.UUID(permission_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid permission_oid "{}"'.format(permission_oid))

        self.log.info('{func}(role_oid={role_oid}, permission_oid={permission_oid}, details={details})',
                      role_oid=hlid(role_oid),
                      permission_oid=hlid(permission_oid),
                      func=hltype(self.get_role_permission),
                      details=details)

        with self.db.begin() as txn:
            role = self.schema.roles[txn, role_oid_]
            if not role:
                raise ApplicationError('crossbar.error.no_such_object', 'no role with oid {} found'.format(role_oid_))

            permission = self.schema.permissions[txn, permission_oid_]
            if not permission:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no arealm permission oid {} found'.format(permission_oid_))

        res_obj = permission.marshal()

        return res_obj

    @wamp.register(None, check_types=True)
    def get_role_permissions_by_uri(self,
                                    role_oid: str,
                                    prefix: Optional[str] = None,
                                    details: Optional[CallDetails] = None) -> List[Dict]:
        """
        Get information for the permission on a role.

        :param role_oid: Object ID of the role to retrieve the permission for.
        :param permission_oid: Object ID of the permission to retrieve.

        :return: Permission definition.
        """
        self.log.info('{func}(role_oid={role_oid}, prefix={prefix}, details={details})',
                      role_oid=hlid(role_oid),
                      prefix=hlval(prefix),
                      func=hltype(self.get_role_permission),
                      details=details)

        try:
            role_oid_ = uuid.UUID(role_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid role_oid "{}"'.format(role_oid))

        with self.db.begin() as txn:
            role = self.schema.roles[txn, role_oid_]
            if not role:
                raise ApplicationError('crossbar.error.no_such_object', 'no role with oid {} found'.format(role_oid_))

            res = []
            from_key = (role_oid_, prefix if prefix else '')
            to_key = (uuid.UUID(int=(int(role_oid_) + 1)), '')
            for permission_oid in self.schema.idx_permissions_by_uri.select(txn,
                                                                            from_key=from_key,
                                                                            to_key=to_key,
                                                                            return_keys=False):
                permission = self.schema.permissions[txn, permission_oid]
                res.append(permission.marshal())

        return res

    @wamp.register(None, check_types=True)
    def list_arealm_roles(self,
                          arealm_oid: str,
                          return_names: Optional[bool] = None,
                          details: Optional[CallDetails] = None) -> List[str]:
        """
        List roles currently associated with the given application realm.

        :param arealm_oid: The application realm to list roles for.

        :return: List of role object IDs of roles associated with the application realm.
        """
        self.log.info('{func}(arealm_oid={arealm_oid}, return_names={return_names}, details={details})',
                      func=hltype(self.list_arealm_roles),
                      arealm_oid=hlid(arealm_oid),
                      return_names=return_names,
                      details=details)

        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(arealm_oid))

        with self.db.begin() as txn:
            arealm = self.schema.arealms[txn, arealm_oid_]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(arealm_oid_))

            res = []
            for _, role_oid in self.schema.arealm_role_associations.select(txn,
                                                                           from_key=(arealm_oid_,
                                                                                     uuid.UUID(bytes=b'\x00' * 16)),
                                                                           to_key=(arealm_oid_,
                                                                                   uuid.UUID(bytes=b'\xff' * 16)),
                                                                           return_values=False):
                if return_names:
                    role = self.schema.roles[txn, role_oid]
                    res.append(role.name)
                else:
                    res.append(str(role_oid))
            return res

    @wamp.register(None, check_types=True)
    async def add_arealm_role(self,
                              arealm_oid: str,
                              role_oid: str,
                              config: Optional[dict] = None,
                              details: Optional[CallDetails] = None) -> dict:
        """
        Add a role to an application realm.

        :param arealm_oid: OID of the application realm to which to add the role.
        :param role_oid: OID of the role to add to the application realm.
            A role can be added to more than one application realm.
        :param config:

        :return:
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(arealm_oid={arealm_oid}, role_oid={role_oid}, config={config}, details={details})',
                      func=hltype(self.add_arealm_role),
                      arealm_oid=hlid(arealm_oid),
                      role_oid=hlid(role_oid),
                      config=config,
                      details=details)

        config = config or {}
        config['arealm_oid'] = arealm_oid
        config['role_oid'] = role_oid
        association = ApplicationRealmRoleAssociation.parse(config)

        with self.db.begin(write=True) as txn:
            arealm = self.schema.arealms[txn, association.arealm_oid]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no application realm with oid {} found'.format(association.arealm_oid))

            role = self.schema.roles[txn, association.role_oid]
            if not role:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no role with oid {} found'.format(association.role_oid))

            self.schema.arealm_role_associations[txn, (association.arealm_oid, association.role_oid)] = association

        res_obj = association.marshal()
        self.log.info('role added to application realm:\n{association}', association=res_obj)

        await self._session.publish('{}.on_arealm_role_added'.format(self._prefix), res_obj, options=self._PUBOPTS)

        return res_obj

    @wamp.register(None, check_types=True)
    async def remove_arealm_role(self, arealm_oid: str, role_oid: str, details: Optional[CallDetails] = None):
        """
        Remove a role from an application realm.

        :param arealm_oid: Object ID of the application realm to remove the role from.
        :param role_oid: Object ID of the role to remove from the application realm.

        :return: Application realm role removal information.
        """
        assert type(arealm_oid) == str
        assert type(role_oid) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(arealm_oid={arealm_oid}, role_oid={role_oid}, details={details})',
                      arealm_oid=hlid(arealm_oid),
                      role_oid=hlid(role_oid),
                      func=hltype(self.remove_arealm_role),
                      details=details)

        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(arealm_oid))

        try:
            role_oid_ = uuid.UUID(role_oid)
        except Exception:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(role_oid))

        with self.db.begin(write=True) as txn:
            arealm = self.schema.arealms[txn, arealm_oid_]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no arealm with oid {} found'.format(arealm_oid_))

            role = self.schema.roles[txn, role_oid_]
            if not role:
                raise ApplicationError('crossbar.error.no_such_object', 'no role with oid {} found'.format(role_oid_))

            arealm_role_association = self.schema.arealm_role_associations[txn, (arealm_oid_, role_oid_)]
            if not arealm_role_association:
                raise ApplicationError(
                    'crossbar.error.no_such_object',
                    'no role association for (arealm_oid={}, role_oid={}) found'.format(arealm_oid_, role_oid_))

            del self.schema.arealm_role_associations[txn, (arealm_oid_, role_oid_)]

        res_obj = arealm_role_association.marshal()
        self.log.info('role removed from arealm:\n{res_obj}', membership=res_obj)

        await self._session.publish('{}.on_arealm_role_removed'.format(self._prefix), res_obj, options=self._PUBOPTS)

        return res_obj

    @wamp.register(None, check_types=True)
    def get_arealm_role(self, arealm_oid: str, role_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Get information for the association of a role with an application realm.

        :param arealm_oid: The application realm for which to return the association for.
        :param role_oid: The role for which to return the association for.

        :return: Application realm role association removal information.
        """
        assert type(arealm_oid) == str
        assert type(role_oid) == str
        assert details is None or isinstance(details, CallDetails)

        try:
            arealm_oid_ = uuid.UUID(arealm_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid arealm_oid "{}"'.format(str(e)))

        try:
            role_oid_ = uuid.UUID(role_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid role_oid "{}"'.format(str(e)))

        self.log.info('{func}(arealm_oid={arealm_oid}, role_oid={role_oid}, details={details})',
                      arealm_oid=hlid(arealm_oid_),
                      role_oid=hlid(role_oid_),
                      func=hltype(self.get_arealm_role),
                      details=details)

        with self.db.begin() as txn:
            arealm = self.schema.arealms[txn, arealm_oid_]
            if not arealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no arealm with oid {} found'.format(arealm_oid_))

            role = self.schema.roles[txn, role_oid_]
            if not role:
                raise ApplicationError('crossbar.error.no_such_object', 'no role with oid {} found'.format(role_oid_))

            association = self.schema.arealm_role_associations[txn, (arealm_oid_, role_oid_)]
            if not association:
                raise ApplicationError(
                    'crossbar.error.no_such_object',
                    'no association between role {} and arealm {} found'.format(arealm_oid, role_oid_))

        res_obj = association.marshal()

        return res_obj



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import txaio

txaio.use_twisted()

from crossbar.network._authenticator import Authenticator as XbrNetworkAuthenticator
from crossbar.network._api import Network as XbrNetwork
from crossbar.network.personality import Personality

__doc__ = Personality.DESC

__all__ = (
    'Personality',
    'XbrNetwork',
    'XbrNetworkAuthenticator',
)
